#!/usr/bin/python3
# -*- mode:python ; tab-width:4 -*- ex:set tabstop=4 shiftwidth=4 expandtab: -*-

#
# YAPySudoku, a Sudoku puzzle game written in Python, with a network mode
# to play with someone else.
# Copyright (C) 2006, 2022 - Nicolas Salguero <nicolas.salguero@laposte.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

__author__ = 'Nicolas Salguero'
__version__ = '$Revision: 148 $'
__date__ = '$Date: 2022-06-10 16:13:15 $'
__copyright__ = 'Copyright (c) 2006, 2022 - Nicolas Salguero '
__copyright__ += '<nicolas.salguero@laposte.net>'
__license__ = 'GPLv3+'


import os
import sys
import wx.adv
import wx.py.frame
import wx.lib.intctrl
import wx.lib.dialogs
import wx.lib.layoutf
import csv
import math
import random
import pygame.mixer
import gettext
import zipfile
import tempfile
import time
import socket
import threading
import configparser
import locale


ID_PAUSE = wx.Window.NewControlId()
ID_SERVER = wx.Window.NewControlId()
ID_JOIN = wx.Window.NewControlId()
ID_SEND = wx.Window.NewControlId()
ID_PORT = wx.Window.NewControlId()
ID_INIT_COLOUR = wx.Window.NewControlId()
ID_X_COLOUR = wx.Window.NewControlId()
ID_BG_COLOUR = wx.Window.NewControlId()
ID_ONE_P_COLOUR = wx.Window.NewControlId()
ID_SERV_P_COLOUR = wx.Window.NewControlId()
ID_CLI_P_COLOUR = wx.Window.NewControlId()
ID_MSG_COLOUR = wx.Window.NewControlId()
ID_RST_COLOURS = wx.Window.NewControlId()
ID_FORCED_UNDO = wx.Window.NewControlId()
ID_STARTUP = wx.Window.NewControlId()

if wx.Platform != '__WXGTK__':
    ID_LICENSE = wx.Window.NewControlId()

DARK_BLUE = wx.Colour(0, 0, 139, 255)
DARK_RED = wx.Colour(139, 0, 0, 255)
DARK_GREEN = wx.Colour(0, 100, 0, 255)
SOME_GREY = wx.Colour(207, 207, 207, 255)

HOST = ''
port = 29979

MIN_PORT = 1024
MAX_PORT = 65536

SCORE_SIZE = 20

def init_colours_with_default_values():
    """Set colours with their default values."""
    global initially_filled_cell_colour, one_player_colour, server_colour, \
           client_colour, msg_colour, x_colour, bg_colour
    initially_filled_cell_colour = DARK_BLUE
    one_player_colour = wx.BLACK
    server_colour = DARK_GREEN
    client_colour = DARK_RED
    msg_colour = wx.BLACK
    x_colour = SOME_GREY
    bg_colour = wx.WHITE

init_colours_with_default_values()


class NoPossibleValueError(Exception):
    """Exception raised when there is no possible value for a cell."""
    def __init__(self, msg):
        self.msg = 'No possible value on cell ' + msg


class NoSolutionError(Exception):
    """Exception raised when the Sudoku puzzle has no solution."""
    def __init__(self, msg):
        self.msg = 'This Sudoku puzzle has no solution!!! ' + msg


class StopError(Exception):
    """Exception raised when the resolution is stopped."""
    def __init__(self, msg):
        self.msg = msg


class SudokuGrid:
    """
    A Sudoku grid is a list of N elements.
    Each element is a list of N integers.

    A Sudoku grid is also logically divided in N square boxes.
    """

    def __init__(self, size_of_side):
        self._size_of_side = size_of_side
        self._size_of_box = int(math.sqrt(self._size_of_side))
        self._blanks_number = self._size_of_side**2
        self._grid = []
        i = 0
        while i < self._size_of_side:
            row = []
            j = 0
            while j < self._size_of_side:
                row.append(None)
                j += 1
            self._grid.append(row)
            i += 1

    def __str__(self):
        result = os.linesep
        for i in range(self._size_of_side):
            if i % self._size_of_box == 0:
                result += '====' * self._size_of_side
                result += '=' + os.linesep
            else:
                result += '----' * self._size_of_side
                result += '-' + os.linesep
            for j in range(self._size_of_side):
                if j % self._size_of_box == 0:
                    result += '| '
                else:
                    result += "' "
                value = self._grid[i][j]
                if value is not None:
                    result += str(value) + ' '
                else:
                    result += '  '
            result += '|' + os.linesep
        result += '====' * self._size_of_side
        result += '=' + os.linesep
        return result

    def Get(self, row_number, column_number):
        """Get the value of the place given in parameter."""
        return self._grid[row_number][column_number]

    def Put(self, value, row_number, column_number):
        """Put the value at the place given in parameter."""
        old_value = self._grid[row_number][column_number]
        self._grid[row_number][column_number] = value
        if value is None and old_value is not None:
            self._blanks_number += 1
        elif value is not None and old_value is None:
            self._blanks_number -= 1

    def Clone(self):
        """Return a deep copy of the grid."""
        other = SudokuGrid(self._size_of_side)
        for i in range(self._size_of_side):
            for j in range(self._size_of_side):
                value = self._grid[i][j]
                if value is not None:
                    other.Put(value, i, j)
        return other

    def GetSizeOfSide(self):
        """Get the size of a side of the Sudoku grid."""
        return self._size_of_side

    def GetSizeOfBox(self):
        """Get the size of a box of the Sudoku grid."""
        return self._size_of_box

    def GetBlanksNumber(self):
        """Get the number of empty places in the Sudoku grid."""
        return self._blanks_number

    def GetRow(self, row_number):
        """Return a row of the Sudoku grid."""
        return self._grid[row_number]


class SudokuHistoryElement:
    """
    A SudokuHistoryElement contains six elements:
      - the coordinates of the modified cell: a tuple of two integers;
      - the old value;
      - the new value;
      - the old colour of the cell;
      - the new colour of the cell.
    """

    def __init__(self, row_number, column_number, old_value, new_value,
                 old_colour, new_colour):
        self._coordinates = (row_number, column_number)
        self._old_value = old_value
        self._new_value = new_value
        self._old_colour = old_colour
        self._new_colour = new_colour

    def __str__(self):
        return '(' + str(self._coordinates) + ', ' + str(self._old_value) + \
               ', ' + str(self._new_value) + ', ' + str(self._old_colour) + \
               ', ' + str(self._new_colour) + ')'

    def GetCoordinates(self):
        """Get the coordinates of the cell."""
        return self._coordinates

    def GetOldValue(self):
        """Get the old value of the cell."""
        return self._old_value

    def GetNewValue(self):
        """Get the new value of the cell."""
        return self._new_value

    def GetOldColour(self):
        """Get the old colour of the cell."""
        return self._old_colour

    def GetNewColour(self):
        """Get the new colour of the cell."""
        return self._new_colour

    def ResetColoursFor1PlayerMode(self, colour):
        """Reset the two colours to the colour used in 1 player mode."""
        self._old_colour = colour
        self._new_colour = colour


class SudokuHistory:
    """
    A pile with all of the modifications in the Sudoku grid.
    There is also an index which indicates the current
    position in the history.
    """

    def __init__(self, sudoku_grid=None, sudoku_frame=None, csv_reader=None,
                 position=None, trace=None):
        if trace is None:
            self._trace = []
        else:
            self._trace = trace
        if position is None:
            self._position = -1
        else:
            self._position = position
        self._sudoku_grid = sudoku_grid
        self._sudoku_frame = sudoku_frame
        if csv_reader is not None:
            self._put_values(csv_reader)

    def __str__(self):
        result = '('
        i = 0
        for elem in self._trace:
            result += str(elem)
            if i < len(self._trace) - 1:
                result += ',' + os.linesep
            i += 1
        result += ')' + os.linesep + 'Current position: ' + str(self._position)
        return result

    def GetPosition(self):
        """Get the position."""
        return self._position

    def GetTrace(self):
        """Get the trace."""
        return self._trace

    def GetCurrentHistoryElement(self):
        """Get the current history element or None if history is empty."""
        if self._position == -1:
            return None
        else:
            return self._trace[self._position]

    def SetSudokuGrid(self, sudoku_grid):
        """Set a new Sudoku grid to the history."""
        self._sudoku_grid = sudoku_grid

    def Push(self, row_number, column_number, old_value=None, new_value=None,
             modify_sudoku_frame=True, old_colour=None, new_colour=None):
        """
        Add a modification in the Sudoku history.

        This modification has to be the current end of the history
        so all the elements of the list after the current position
        in the history have to be removed.
        """
        self._set_last_modification_to_position()
        self._trace.append(SudokuHistoryElement(row_number, column_number,
                                                old_value, new_value,
                                                old_colour, new_colour))
        self._position += 1
        if self._sudoku_grid is not None:
            self._sudoku_grid.Put(new_value, row_number, column_number)
        if modify_sudoku_frame and self._sudoku_frame is not None:
            self._sudoku_frame.SetValue(new_value, row_number, column_number)

    def MoveBack(self, definitive=False):
        """
        Make an 'Undo'. If it is a 'definitive Undo', there is no
        possibility to do a 'Redo' after that.

        Return True if the operation succeeded.
        """
        if self._position > -1:
            row_number, column_number = \
                    self._trace[self._position].GetCoordinates()
            old_value = self._trace[self._position].GetOldValue()
            old_colour = self._trace[self._position].GetOldColour()
            if definitive:
                self._remove()
            self._position -= 1
            if self._sudoku_grid is not None:
                self._sudoku_grid.Put(old_value, row_number, column_number)
            if self._sudoku_frame is not None:
                self._sudoku_frame.SetValue(old_value, row_number,
                                            column_number, colour=old_colour)
            return True
        else:
            return False

    def MoveOn(self):
        """
        Make a 'Redo'.

        Return True if the operation succeeded.
        """
        if self._position < len(self._trace) - 1:
            self._position += 1
            row_number, column_number = \
                                   self._trace[self._position].GetCoordinates()
            new_value = self._trace[self._position].GetNewValue()
            if self._sudoku_grid is not None:
                self._sudoku_grid.Put(new_value, row_number, column_number)
            if self._sudoku_frame is not None:
                new_colour=self._trace[self._position].GetNewColour()
                self._sudoku_frame.SetValue(new_value, row_number,
                                            column_number, colour=new_colour)
            return True
        else:
            return False

    def Clear(self, grid_too=False):
        """Totally clear the history and, optionally, the grid too."""
        if grid_too:
            self._set_last_modification_to_position()
            while self.MoveBack(True):
                pass
        else:
            self._trace = []
            self._position = -1

    def ResetColourFor1PlayerMode(self, colour):
        """Reset the colour to the colour used in 1 player mode."""
        for elem in self._trace:
            elem.ResetColoursFor1PlayerMode(colour)

    def _set_last_modification_to_position(self):
        """Remove all modifications after the current position."""
        if self._position < len(self._trace) - 1:
            for position in range(len(self._trace) - 1, self._position, -1):
                self._remove(position)

    def _remove(self, position=None):
        """
        Remove a modification in the Sudoku history.

        If no position is given, remove the last modification
        recorded in the history.
        """
        if position is not None:
            del(self._trace[position])
        else:
            del(self._trace[self._position])

    def _put_values(self, csv_reader):
        """Initialise the grid with a CSV reader."""
        i = 0
        for row in csv_reader:
            for j in range(self._sudoku_grid.GetSizeOfSide()):
                value = row[j]
                if value != '':
                    self._sudoku_grid.Put(int(value), i, j)
                    self._sudoku_frame.SetValue(int(value), i, j)
            i += 1


class SudokuTry:
    """
    A try contains three elements:
      - the coordinates of a cell: a tuple of two integers;
      - the possible values for that cell: a list of integers;
      - a SudokuHistory object containing the modifications in the
        Sudoku grid which are due to testing one of the possible
        values for the given cell.
    """

    def __init__(self, row_number, column_number, possible_values):
        self._coordinates = (row_number, column_number)
        self._possible_values = possible_values
        self._history = SudokuHistory()

    def __str__(self):
        return '(' + str(self._coordinates) + ', ' + \
                str(self._possible_values) + ', ' + \
                str(self._history) + ')'

    def GetCoordinates(self):
        """Get the coordinates of the cell."""
        return self._coordinates

    def GetHistory(self):
        """Get the history of the try."""
        return self._history

    def GetFirstPossibleValue(self):
        """Get the first possible value for the cell."""
        return self._possible_values[0]

    def RemoveFirstPossibleValue(self):
        """Remove the first possible value for the cell."""
        self._possible_values.pop(0)

    def HasNoMorePossibleValues(self):
        """Return True if the cell has no more possible values."""
        return self._possible_values == []


class SudokuTries:
    """The list of tries, created during the resolution of the Sudoku."""

    def __init__(self):
        self._tries = []

    def __str__(self):
        result = '('
        i = 0
        for sudoku_try in self._tries:
            result += str(sudoku_try)
            if i < len(self._tries) - 1:
                result += ',' + os.linesep
            i += 1
        result += ')' + os.linesep
        return result

    def Add(self, row_number, column_number, possible_values):
        """Add a try to the list of tries."""
        self._tries.append(SudokuTry(row_number, column_number,
                                     possible_values))

    def GetCurrentTry(self):
        """Return the current try, which is the last one."""
        return self._tries[len(self._tries) - 1]

    def RemoveCurrentTry(self):
        """Remove the current try."""
        del(self._tries[len(self._tries) - 1])

    def NoMore(self):
        """Return True if there is only one try in the list."""
        return len(self._tries) == 1


class SudokuPuzzle:
    """
    The purpose of this class is to try to solve a Sudoku puzzle.
    The constraints are described in this class.
    """

    def __init__(self, sudoku_frame, size_of_sudoku, x_sudoku,
                 csv_reader=None, position=None, trace=None):
        self._x_sudoku = x_sudoku
        self._stop_resolution = False
        self._modify_sudoku_frame = True
        self._sudoku_grid = SudokuGrid(size_of_sudoku)
        self._history = SudokuHistory(self._sudoku_grid, sudoku_frame,
                                      csv_reader, position, trace)

    def __str__(self):
        return str(self._sudoku_grid)

    def GetSudokuGrid(self):
        """Return the Sudoku grid."""
        return self._sudoku_grid

    def GetHistory(self):
        """Return the history."""
        return self._history

    def Solve(self):
        """Solve the sudoku."""
        if not self._is_initially_good():
            raise NoSolutionError(
                    'The Sudoku puzzle does not initially respect the rules.')
        progression = True
        bad_choice = trying_value = False
        tries = SudokuTries()
        while self._sudoku_grid.GetBlanksNumber() != 0:
            if progression:
                progression = False
            else:
                trying_value = True
                if bad_choice:
                    bad_choice = False
                    self._backtracking(tries)
                else:
                    row_number, column_number = self._get_best_blank()
                    possible_values = self._get_possible_values(row_number,
                                                                column_number)
                    tries.Add(row_number, column_number, possible_values)
                possible_value = tries.GetCurrentTry().GetFirstPossibleValue()
                row_number, column_number = \
                                        tries.GetCurrentTry().GetCoordinates()
                self._set_value(possible_value, row_number, column_number)
            progression, bad_choice = self._grid_travel(tries, trying_value,
                                                        progression,
                                                        bad_choice)

    def CreateNewGame(self, difficulty):
        """Create a new Sudoku grid with some random values."""
        bad_grid = True
        while bad_grid:
            try:
                self.Solve()
                bad_grid = False
            except NoSolutionError:
                self._history.Clear(True)
                print('[FIXME] a solution must exist')
                print(self)
        self._randomly_remove_values(difficulty)
        self._history.Clear()

    def Stop(self):
        """Stop the search of solution."""
        self._stop_resolution = True

    def IsAPossibleValue(self, value, row_number, column_number):
        """
        Return True if the value given in parameter is a possible
        value for the cell given in parameter.
        """
        old_value = self._sudoku_grid.Get(row_number, column_number)
        self._sudoku_grid.Put(value, row_number, column_number)
        result = self._is_a_good_row(row_number) and \
                 self._are_good_boxes_on_row(row_number, column_number) and \
                 self._is_a_good_column(column_number) and \
                 self._are_good_boxes_on_column(row_number, column_number) and \
                 self._is_a_good_box(row_number, column_number)
        if result and self._x_sudoku:
            if row_number == column_number:
                result = self._is_a_good_first_diag() and \
                         self._are_good_boxes_on_first_diag(row_number,
                                                            column_number)
            if row_number + column_number + 1 == \
                                             self._sudoku_grid.GetSizeOfSide():
                result = result and self._is_a_good_second_diag() and \
                         self._are_good_boxes_on_second_diag(row_number,
                                                             column_number)
        self._sudoku_grid.Put(old_value, row_number, column_number)
        return result

    def _get_box_number(self, row_number, column_number):
        """Return the box that contains the coordinates given in parameter."""
        return (row_number // self._sudoku_grid.GetSizeOfBox()) * \
                                           self._sudoku_grid.GetSizeOfBox() + \
               column_number // self._sudoku_grid.GetSizeOfBox()

    def _get_other_boxes_of_row(self, row_number, column_number):
        """
        Return the boxes of a row that does not contain the coordinates
        given in parameter.
        """
        box_number = self._get_box_number(row_number, column_number)
        result = []
        for i in range(self._sudoku_grid.GetSizeOfBox()):
            number = (box_number // self._sudoku_grid.GetSizeOfBox()) * \
                                           self._sudoku_grid.GetSizeOfBox() + i
            if number != box_number:
                result.append(number)
        return result

    def _get_other_boxes_of_column(self, row_number, column_number):
        """
        Return the boxes of a column that does not contain the coordinates
        given in parameter.
        """
        box_number = self._get_box_number(row_number, column_number)
        result = []
        for i in range(1, self._sudoku_grid.GetSizeOfBox()):
            number = (box_number + i * self._sudoku_grid.GetSizeOfBox()) % \
                                              self._sudoku_grid.GetSizeOfSide()
            result.append(number)
        return result

    def _get_other_boxes_of_first_diag(self, row_number, column_number):
        """
        Return the boxes of the first diagonal that does not contain
        the coordinates given in parameter.
        """
        box_number = self._get_box_number(row_number, column_number)
        result = []
        number = 0
        while number < self._sudoku_grid.GetSizeOfSide():
            if number != box_number:
                result.append(number)
            number += self._sudoku_grid.GetSizeOfBox() + 1
        return result

    def _get_other_boxes_of_second_diag(self, row_number, column_number):
        """
        Return the boxes of the second_diagonal that does not contain
        the coordinates given in parameter.
        """
        box_number = self._get_box_number(row_number, column_number)
        result = []
        number = self._sudoku_grid.GetSizeOfBox() * \
                                        (self._sudoku_grid.GetSizeOfBox() - 1)
        while number >= self._sudoku_grid.GetSizeOfBox() - 1:
            if number != box_number:
                result.append(number)
            number -= self._sudoku_grid.GetSizeOfBox() - 1
        return result

    def _is_a_possible_box(self, box_number):
        """Test if a box respects the rules."""
        blanks, values = self._get_blanks_and_values_of_box(box_number)
        possible_values = self._get_possible_values_of_box(blanks)
        if possible_values is None:
            return False
        values.extend(possible_values)
        values.sort()
        available_values = set(values)
        all_values = set(range(1, self._sudoku_grid.GetSizeOfSide() + 1))
        return (available_values == all_values)

    def _are_good_boxes_on_first_diag(self, row_number, column_number):
        """
        Test if the other boxes across the first diagonal respect the rules.
        """
        other_boxes = self._get_other_boxes_of_first_diag(row_number,
                                                          column_number)
        for box_number in other_boxes:
            if not self._is_a_possible_box(box_number):
                return False
        return True

    def _are_good_boxes_on_second_diag(self, row_number, column_number):
        """
        Test if the other boxes across the second diagonal respect the rules.
        """
        other_boxes = self._get_other_boxes_of_second_diag(row_number,
                                                           column_number)
        for box_number in other_boxes:
            if not self._is_a_possible_box(box_number):
                return False
        return True

    def _are_good_boxes_on_row(self, row_number, column_number):
        """Test if the other boxes across a row respect the rules."""
        other_boxes = self._get_other_boxes_of_row(row_number, column_number)
        for box_number in other_boxes:
            if not self._is_a_possible_box(box_number):
                return False
        return True

    def _are_good_boxes_on_column(self, row_number, column_number):
        """Test if the other boxes across a column respect the rules."""
        other_boxes = self._get_other_boxes_of_column(row_number, column_number)
        for box_number in other_boxes:
            if not self._is_a_possible_box(box_number):
                return False
        return True

    def _ensure_unique_solution(self, possible_values, row_number,
                                column_number, value):
        """
        Be sure that the Sudoku grid has a unique solution
        by trying to solve other Sudokus. In those ones,
        the cell given in parameter has a value which is possible
        but different from the one given in parameter.
        """
        saved_grid = self._sudoku_grid.Clone()
        self._modify_sudoku_frame = False
        unique_solution = True
        for val in possible_values:
            self._sudoku_grid.Put(val, row_number, column_number)
            try:
                self.Solve()
                unique_solution = False
                break
            except NoSolutionError:
                if val != possible_values[len(possible_values) - 1]:
                    self._sudoku_grid = saved_grid.Clone()
                    self._history.SetSudokuGrid(self._sudoku_grid)
        self._sudoku_grid = saved_grid
        self._history.SetSudokuGrid(self._sudoku_grid)
        self._modify_sudoku_frame = True
        if not unique_solution:
            self._set_value(value, row_number, column_number)

    def _try_to_remove_a_value(self, row_number, column_number):
        """Try to remove a value from a Sudoku grid."""
        value = self._sudoku_grid.Get(row_number, column_number)
        self._set_value(None, row_number, column_number)
        possible_values = [value]
        try:
            possible_values = self._get_possible_values(row_number,
                                                        column_number)
        except NoPossibleValueError:
            self._set_value(value, row_number, column_number)
            print('[FIXME] value = ' + str(value))
            print('row = ' + str(row_number + 1))
            print('column = ' + str(column_number + 1))
            print(self)
        possible_values.remove(value)
        if possible_values != []:
            self._ensure_unique_solution(possible_values, row_number,
                                         column_number, value)

    def _randomly_remove_values(self, difficulty):
        """Randomly remove some values from a complete Sudoku grid."""
        coords_list = []
        for i in range(self._sudoku_grid.GetSizeOfSide()):
            for j in range(self._sudoku_grid.GetSizeOfSide()):
                coords_list.append([i, j])
        random.shuffle(coords_list)
        max_number = self._sudoku_grid.GetSizeOfSide()**2
        max_number -= self._sudoku_grid.GetSizeOfSide()**2 // \
                                               self._sudoku_grid.GetSizeOfBox()
        if self._x_sudoku:
            max_number += self._sudoku_grid.GetSizeOfBox()
            if difficulty == 3:
                max_number += self._sudoku_grid.GetSizeOfBox()
        if difficulty == 0:
            max_number -= self._sudoku_grid.GetSizeOfSide()
        elif difficulty == 1:
            max_number -= self._sudoku_grid.GetSizeOfBox()
        for coords in coords_list:
            if (difficulty < 3 or self._x_sudoku) and \
                            self._sudoku_grid.GetBlanksNumber() > max_number:
                break
            self._try_to_remove_a_value(coords[0], coords[1])

    def _is_initially_good(self):
        """Test if the Sudoku puzzle respects the rules at the beginning."""
        for box_number in range(self._sudoku_grid.GetSizeOfSide()):
            if not self._is_a_possible_box(box_number):
                return False
        return True

    def _get_first_cell_of_box(self, box_number):
        """
        Return the coordinates of the first cell
        of the box given in parameter.
        """
        column_number = (box_number % self._sudoku_grid.GetSizeOfBox()) * \
                        self._sudoku_grid.GetSizeOfBox()
        row_number = (box_number // self._sudoku_grid.GetSizeOfBox()) * \
                        self._sudoku_grid.GetSizeOfBox()
        return [row_number, column_number]

    def _get_blanks_and_values_of_box(self, box_number):
        """Return all the blanks of the box given in parameter."""
        row_number, column_number = self._get_first_cell_of_box(box_number)
        blanks = []
        values = []
        for i in range(row_number, row_number +
                                   self._sudoku_grid.GetSizeOfBox()):
            for j in range(column_number, column_number +
                                          self._sudoku_grid.GetSizeOfBox()):
                val = self._sudoku_grid.Get(i, j)
                if val is None:
                    blanks.append([i, j])
                else:
                    values.append(val)
        return [blanks, values]

    def _get_possible_values_of_box(self, blanks):
        """Return the possible values of the box given in parameter."""
        possible_values = []
        for blank in blanks:
            try:
                values = self._get_possible_values(blank[0], blank[1])
                is_sure = False
                for value in values:
                    val = self._get_sure_value(blank[0], blank[1], value)
                    sure = (val is not None)
                    if sure:
                        if is_sure:
                            return None
                        else:
                            is_sure = sure
            except NoPossibleValueError:
                return None
            possible_values.extend(values)
        return possible_values

    def _grid_travel(self, tries, trying_value, progression, bad_choice):
        """Travel the sudoku grid to add sure values."""
        try:
            for i in range(self._sudoku_grid.GetSizeOfSide()):
                for j in range(self._sudoku_grid.GetSizeOfSide()):
                    if self._sudoku_grid.Get(i, j) is None:
                        sure_value = self._get_sure_value(i, j)
                        if sure_value is not None:
                            self._set_value(sure_value, i, j)
                            progression = True
                            if trying_value:
                                tries.GetCurrentTry().GetHistory().Push(i, j)
        except NoPossibleValueError as npve:
            if trying_value:
                progression = False
                bad_choice = True
            else:
                raise NoSolutionError(npve.msg)
        return [progression, bad_choice]

    def _backtracking(self, tries):
        """Do backtracking to try another possible value."""
        backtracking = True
        while backtracking:
            backtracking = False
            tries.GetCurrentTry().RemoveFirstPossibleValue()
            historyElem = \
                  tries.GetCurrentTry().GetHistory().GetCurrentHistoryElement()
            while historyElem is not None:
                row_number, column_number = historyElem.GetCoordinates()
                self._set_value(None, row_number, column_number)
                tries.GetCurrentTry().GetHistory().MoveBack(True)
                historyElem = \
                  tries.GetCurrentTry().GetHistory().GetCurrentHistoryElement()
            if tries.GetCurrentTry().HasNoMorePossibleValues():
                if tries.NoMore():
                    raise NoSolutionError(
                             'None of the possible values for ' +
                             str(tries.GetCurrentTry()) +
                             ' lead to the solution.')
                else:
                    row_number, column_number = \
                                         tries.GetCurrentTry().GetCoordinates()
                    self._set_value(None, row_number, column_number)
                    tries.RemoveCurrentTry()
                    backtracking = True

    def _set_value(self, value, row_number, column_number):
        """Set the value at the position given in parameter."""
        if self._stop_resolution:
            self._stop_resolution = False
            raise StopError('Stop the search of solution.')
        old_value = self._sudoku_grid.Get(row_number, column_number)
        self._history.Push(row_number, column_number, old_value, value,
                           self._modify_sudoku_frame)

    def _get_best_blank(self):
        """
        Return one of the blanks in the sudoku which
        have the smallest number of possible values.
        """
        best_blank = []
        num_best_poss_val = sys.maxsize
        for i in range(self._sudoku_grid.GetSizeOfSide()):
            for j in range(self._sudoku_grid.GetSizeOfSide()):
                if self._sudoku_grid.Get(i, j) is None:
                    num_poss_val = len(self._get_possible_values(i, j))
                    if num_poss_val == 2:
                        return [i, j]
                    if num_poss_val < num_best_poss_val:
                        best_blank = [i, j]
                        num_best_poss_val = num_poss_val
        return best_blank

    def _is_a_good_value(self, values):
        """
        Check a list of values to see if there is not
        twice the same number.

        Return True if the list respects the sudoku rules.
        """
        return len(values) == len(set(values))

    def _is_a_good_first_diag(self):
        """
        Check the first diagonal to see if there is not twice
        the same number.

        Return True if the diagnal respects the sudoku rules.
        """
        values = []
        for i in range(self._sudoku_grid.GetSizeOfSide()):
            val = self._sudoku_grid.Get(i, i)
            if val is not None:
                values.append(val)
        return self._is_a_good_value(values)

    def _is_a_good_second_diag(self):
        """
        Check the second diagonal to see if there is not twice
        the same number.

        Return True if the diagonal respects the sudoku rules.
        """
        values = []
        for i in range(self._sudoku_grid.GetSizeOfSide()):
            j = self._sudoku_grid.GetSizeOfSide() - i - 1
            val = self._sudoku_grid.Get(i, j)
            if val is not None:
                values.append(val)
        return self._is_a_good_value(values)

    def _is_a_good_row(self, row_number):
        """
        Check a row to see if there is not twice
        the same number.

        Return True if the row respects the sudoku rules.
        """
        values = []
        for i in range(self._sudoku_grid.GetSizeOfSide()):
            val = self._sudoku_grid.Get(row_number, i)
            if val is not None:
                values.append(val)
        return self._is_a_good_value(values)

    def _is_a_good_column(self, column_number):
        """
        Check a column to see if there is not twice
        the same number.

        Return True if the column respects the sudoku rules.
        """
        values = []
        for i in range(self._sudoku_grid.GetSizeOfSide()):
            val = self._sudoku_grid.Get(i, column_number)
            if val is not None:
                values.append(val)
        return self._is_a_good_value(values)

    def _is_a_good_box(self, row_number, column_number):
        """
        Check the box that contains the coordinates given in parameter
        to see if there is not twice the same number.

        Return True if the box respects the sudoku rules.
        """
        colnum = (column_number // self._sudoku_grid.GetSizeOfBox()) * \
                 self._sudoku_grid.GetSizeOfBox()
        rownum = (row_number // self._sudoku_grid.GetSizeOfBox()) * \
                 self._sudoku_grid.GetSizeOfBox()
        values = []
        for i in range(rownum, rownum + self._sudoku_grid.GetSizeOfBox()):
            for j in range(colnum, colnum + self._sudoku_grid.GetSizeOfBox()):
                val = self._sudoku_grid.Get(i, j)
                if val is not None:
                    values.append(val)
        return self._is_a_good_value(values)

    def _get_values_with_diag(self, random_values, row_number, column_number):
        """Return a first list of values for the cell given in parameter."""
        values = random_values
        if row_number == column_number or \
           row_number + column_number + 1 == self._sudoku_grid.GetSizeOfSide():
            v = []
            for val in values:
                self._sudoku_grid.Put(val, row_number, column_number)
                good = True
                if row_number == column_number:
                    good = self._is_a_good_first_diag()
                if row_number + column_number + 1 == \
                                             self._sudoku_grid.GetSizeOfSide():
                    good = good and self._is_a_good_second_diag()
                if good:
                    v.append(val)
            self._sudoku_grid.Put(None, row_number, column_number)
            values = v
        if values == []:
            raise NoPossibleValueError('(' + str(row_number + 1) + ', ' +
                                           str(column_number + 1) + ')')
        return values

    def _get_random_values(self):
        """Return all the values in random order."""
        values = list(range(1, self._sudoku_grid.GetSizeOfSide() + 1))
        random.shuffle(values)
        return values

    def _get_possible_values(self, row_number, column_number):
        """Return the possible values for the cell given in parameter."""
        possible_values = []
        values = self._get_random_values()
        if self._x_sudoku:
            values = self._get_values_with_diag(values, row_number,
                                                column_number)
        for val in values:
            self._sudoku_grid.Put(val, row_number, column_number)
            if self._is_a_good_row(row_number) and \
               self._is_a_good_column(column_number) and \
               self._is_a_good_box(row_number, column_number):
                possible_values.append(val)
        self._sudoku_grid.Put(None, row_number, column_number)
        if possible_values == []:
            raise NoPossibleValueError('(' + str(row_number + 1) + ', ' +
                                           str(column_number + 1) + ')')
        return possible_values

    def _get_other_blanks_of_row(self, row_number, column_number):
        """
        Return the other blanks of the row that contains the cell
        given in parameter.
        """
        blanks = []
        for j in range(self._sudoku_grid.GetSizeOfSide()):
            if j != column_number and \
               self._sudoku_grid.Get(row_number, j) is None:
                blanks.append([row_number, j])
        return blanks

    def _get_other_blanks_of_column(self, row_number, column_number):
        """
        Return the other blanks of the column that contains the cell
        given in parameter.
        """
        blanks = []
        for i in range(self._sudoku_grid.GetSizeOfSide()):
            if i != row_number and \
               self._sudoku_grid.Get(i, column_number) is None:
                blanks.append([i, column_number])
        return blanks

    def _get_other_blanks_of_box(self, row_number, column_number):
        """
        Return the other blanks of the box that contains the cell
        given in parameter.
        """
        colnum = (column_number // self._sudoku_grid.GetSizeOfBox()) * \
                 self._sudoku_grid.GetSizeOfBox()
        rownum = (row_number // self._sudoku_grid.GetSizeOfBox()) * \
                 self._sudoku_grid.GetSizeOfBox()
        blanks = []
        for i in range(rownum, rownum + self._sudoku_grid.GetSizeOfBox()):
            for j in range(colnum, colnum + self._sudoku_grid.GetSizeOfBox()):
                if (i != row_number or j != column_number) and \
                   self._sudoku_grid.Get(i, j) is None:
                    blanks.append([i, j])
        return blanks

    def _get_other_blanks_of_first_diag(self, row_number):
        """
        Return the blanks of the first diagonal
        other than the one given in parameter.
        """
        blanks = []
        for i in range(self._sudoku_grid.GetSizeOfSide()):
            if i != row_number and self._sudoku_grid.Get(i, i) is None:
                blanks.append([i, i])
        return blanks

    def _get_other_blanks_of_second_diag(self, row_number):
        """
        Return the blanks of the second diagonal
        other than the one given in parameter.
        """
        blanks = []
        for i in range(self._sudoku_grid.GetSizeOfSide()):
            j = self._sudoku_grid.GetSizeOfSide() - i - 1
            if i != row_number and self._sudoku_grid.Get(i, j) is None:
                blanks.append([i, j])
        return blanks

    def _get_sure_value_of_row(self, possible_values, row_number,
                               column_number):
        """
        Return the mandatory value for the place given in parameter
        or None if no value is sure.

        In this case, there is no possible other place in the row
        for the value because of neighbourhood constraints.
        """
        other_blanks_of_row = self._get_other_blanks_of_row(row_number,
                                                            column_number)
        for val in possible_values:
            sure = True
            for blank in other_blanks_of_row:
                self._sudoku_grid.Put(val, blank[0], blank[1])
                sure = not (self._is_a_good_box(blank[0], blank[1]) and \
                            self._is_a_good_column(blank[1]))
                if self._x_sudoku and blank[0] == blank[1]:
                    sure = sure or not self._is_a_good_first_diag()
                if self._x_sudoku and \
                   blank[1] == self._sudoku_grid.GetSizeOfSide() - blank[0] - 1:
                    sure = sure or not self._is_a_good_second_diag()
                self._sudoku_grid.Put(None, blank[0], blank[1])
                if not sure:
                    break
            if sure:
                return val
        return None

    def _get_sure_value_of_column(self, possible_values, row_number,
                                  column_number):
        """
        Return the mandatory value for the place given in parameter
        or None if no value is sure.

        In this case, there is no possible other place in the column
        for the value because of neighbourhood constraints.
        """
        other_blanks_of_col = self._get_other_blanks_of_column(row_number,
                                                               column_number)
        for val in possible_values:
            sure = True
            for blank in other_blanks_of_col:
                self._sudoku_grid.Put(val, blank[0], blank[1])
                sure = not (self._is_a_good_row(blank[0]) and \
                            self._is_a_good_box(blank[0], blank[1]))
                if self._x_sudoku and blank[0] == blank[1]:
                    sure = sure or not self._is_a_good_first_diag()
                if self._x_sudoku and \
                   blank[1] == self._sudoku_grid.GetSizeOfSide() - blank[0] - 1:
                    sure = sure or not self._is_a_good_second_diag()
                self._sudoku_grid.Put(None, blank[0], blank[1])
                if not sure:
                    break
            if sure:
                return val
        return None

    def _get_sure_value_of_box(self, possible_values, row_number,
                               column_number):
        """
        Return the mandatory value for the place given in parameter
        or None if no value is sure.

        In this case, there is no possible other place in the box
        for the value because of neighbourhood constraints.
        """
        other_blanks_of_box = self._get_other_blanks_of_box(row_number,
                                                            column_number)
        for val in possible_values:
            sure = True
            for blank in other_blanks_of_box:
                self._sudoku_grid.Put(val, blank[0], blank[1])
                sure = not (self._is_a_good_row(blank[0]) and \
                            self._is_a_good_column(blank[1]))
                if self._x_sudoku and blank[0] == blank[1]:
                    sure = sure or not self._is_a_good_first_diag()
                if self._x_sudoku and \
                   blank[1] == self._sudoku_grid.GetSizeOfSide() - blank[0] - 1:
                    sure = sure or not self._is_a_good_second_diag()
                self._sudoku_grid.Put(None, blank[0], blank[1])
                if not sure:
                    break
            if sure:
                return val
        return None

    def _get_sure_value_of_first_diag(self, possible_values, row_number):
        """
        Return the mandatory value for the place given in parameter
        or None if no value is sure.

        In this case, there is no possible other place in the first
        diagonal for the value because of neighbourhood constraints.
        """
        other_blanks_of_diag = self._get_other_blanks_of_first_diag(row_number)
        for val in possible_values:
            sure = True
            for blank in other_blanks_of_diag:
                self._sudoku_grid.Put(val, blank[0], blank[1])
                sure = not (self._is_a_good_first_diag() and \
                            self._is_a_good_row(blank[0]) and \
                            self._is_a_good_column(blank[1]) and \
                            self._is_a_good_box(blank[0], blank[1]))
                self._sudoku_grid.Put(None, blank[0], blank[1])
                if not sure:
                    break
            if sure:
                return val
        return None

    def _get_sure_value_of_second_diag(self, possible_values, row_number):
        """
        Return the mandatory value for the place given in parameter
        or None if no value is sure.

        In this case, there is no possible other place in the second
        diagonal for the value because of neighbourhood constraints.
        """
        other_blanks_of_diag = self._get_other_blanks_of_second_diag(row_number)
        for val in possible_values:
            sure = True
            for blank in other_blanks_of_diag:
                self._sudoku_grid.Put(val, blank[0], blank[1])
                sure = not (self._is_a_good_second_diag() and \
                            self._is_a_good_row(blank[0]) and \
                            self._is_a_good_column(blank[1]) and \
                            self._is_a_good_box(blank[0], blank[1]))
                self._sudoku_grid.Put(None, blank[0], blank[1])
                if not sure:
                    break
            if sure:
                return val
        return None

    def _get_sure_value(self, row_number, column_number, value=None):
        """
        Return the mandatory value for the place given in parameter
        or None if no value is sure.
        """
        if value is None:
            possible_values = self._get_possible_values(row_number, column_number)
            if len(possible_values) == 1:
                return possible_values[0]
        else:
            possible_values = [value]
        if self._x_sudoku:
            val = None
            if row_number == column_number:
                val = self._get_sure_value_of_first_diag(possible_values,
                                                         row_number)
            if val is None and row_number + column_number + 1 == \
                                            self._sudoku_grid.GetSizeOfSide():
                val = self._get_sure_value_of_second_diag(possible_values,
                                                          row_number)
            if val is not None:
                return val
        val = self._get_sure_value_of_row(possible_values, row_number,
                                          column_number)
        if val is not None:
            return val
        val = self._get_sure_value_of_column(possible_values, row_number,
                                             column_number)
        if val is not None:
            return val
        val = self._get_sure_value_of_box(possible_values, row_number,
                                          column_number)
        return val


class SudokuClient(threading.Thread):
    """Sudoku client."""

    def __init__(self, sudoku_frame, server):
        threading.Thread.__init__(self)
        self._server = server
        self._sudoku_frame = sudoku_frame
        self._conn = None
        self._your_time = None
        self._other_player_time = None
        self._finished = threading.Event()
        self._peer_responded = threading.Event()
        self._peer_ok = True
        self._lock = threading.RLock()
        self._peer_response_lock = threading.Lock()

    def stop(self):
        self._finished.set()
        self._lock.acquire()
        if self._conn is not None:
            self._stop_waiting_for_data()
        self._lock.release()
        self.join()

    def run(self):
        self._open_socket()
        if self._conn is not None:
            self._lock.release()
            while 1:
                data = self._read_from_socket()
                self._lock.acquire()
                if not data or data == 'QUIT':
                    self._lock.release()
                    break
                handle_ok = self._handle(data)
                self._lock.release()
                if not handle_ok:
                    break
            self._close_socket()
            self._end_of_connection()
        else:
            self._lock.release()

    def IsPeerOk(self):
        """Indicate whether the other player accepted your command."""
        result = None
        while result is None:
            self._peer_responded.wait(5.0)
            self._peer_response_lock.acquire()
            if not self._peer_responded.isSet():
                self._lock.acquire()
                if self._conn is None:
                    result = True
                self._lock.release()
            else:
                result = self._peer_ok
                self._peer_responded.clear()
            self._peer_response_lock.release()
        return result

    def WriteIntoSocket(self, cmd, arg):
        """
        Write some data into the socket.

        Return the size of the written data.
        """
        self._lock.acquire()
        result = None
        if self._conn is not None:
            str_arg = str(arg)
            data = cmd + ' ' + str(len(bytes(str_arg, 'utf-8'))) + ' ' + str_arg
            try:
                result = self._conn.send(data.encode('utf-8'))
            except socket.error:
                result = None
        self._lock.release()
        return result

    def _stop_waiting_for_data(self):
        """Stop waiting for data."""
        if os.name == 'posix':
            self._conn.shutdown(socket.SHUT_RDWR)
        else:
            try:
                self._conn.send('QUIT')
            except socket.error:
                pass

    def _read_from_socket(self, data_size=4):
        """
        Read some data from the socket.

        Return the data.
        """
        result = None
        if self._conn is not None:
            try:
                result = self._conn.recv(data_size).decode('utf-8')
            except socket.error:
                result = None
            except AttributeError:
                result = None
        return result

    def _send_local_time(self):
        """Send the local time to the other player."""
        self._your_time = time.time()
        self.WriteIntoSocket('CLK', self._your_time)

    def _open_socket(self):
        """Connect to a server."""
        self._lock.acquire()
        self._conn = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            self._conn.connect((self._server, port))
            self._send_local_time()
            wx.CallAfter(self._sudoku_frame.ResetGrid)
            wx.CallAfter(self._sudoku_frame.DisplayMessage,
                         _('Joined the server ') + self._server, msg_colour)
        except socket.error:
            self._close_socket()
            self._sudoku_frame.NetworkProblemOccurred(_('Cannot join server ')
                                                      + self._server + _('!'))

    def _close_socket(self):
        """Close the socket."""
        self._lock.acquire()
        self._conn.close()
        self._conn = None
        self._lock.release()

    def _handle(self, cmd):
        """
        Respond to received data.

        Return False if an error occurred while reading data from the socket.
        """
        msg_size = ''
        data = self._read_from_socket(1)
        if data is None:
            return False
        while data != ' ':
            msg_size += data
            data = self._read_from_socket(1)
        msg_size = int(msg_size)
        data = self._read_from_socket(msg_size)
        if data is None:
            return False
        if cmd == 'MSG ':
            self._handle_message(data)
        elif cmd == 'CLK ':
            self._handle_clock(data)
        elif cmd == 'CHR ':
            self._handle_entered_char(data)
        elif cmd == 'VAL ':
            self._handle_value(data)
        elif cmd == 'RSP ':
            self._handle_response(data)
        elif cmd == 'NEW ':
            self._handle_new_game(data)
        elif cmd == 'DST ':
            self._handle_destroy_grid()
        elif cmd == 'CRT ':
            self._handle_create_grid(data)
        elif cmd == 'DIF ':
            self._handle_difficulty(data)
        elif cmd == 'EOC ':
            self._handle_end_of_creation()
        elif cmd == 'PSE ':
            self._handle_pause_game(data)
        elif cmd == 'UND ':
            self._handle_undo()
        elif cmd == 'RED ':
            self._handle_redo()
        elif cmd == 'RET ':
            self._handle_retry(data)
        elif cmd == 'CLR ':
            self._handle_recolour(data)
        elif cmd == 'AFU ':
            self._handle_allow_forcing_undo(data)
        return True

    def _handle_message(self, data):
        """Manage a message received from the other player."""
        wx.CallAfter(self._sudoku_frame.DisplayMessage, data,
                     self._sudoku_frame.GetOtherPlayerColour())

    def _handle_clock(self, data):
        """Set the remote time."""
        self._other_player_time = float(data)

    def _handle_entered_char(self, data):
        """
        Indicate whether the other player had the right to try
        to put a value in a cell.
        """
        time, row_number, column_number = data.strip('[]').split(',')
        time = float(time)
        row_number, column_number = int(row_number), int(column_number)
        if self._is_peer_ask_ok(time):
            self.WriteIntoSocket('RSP', 'OK')
            self._sudoku_frame.GetLock().acquire()
            colour = self._sudoku_frame.GetOtherPlayerColour()
            wx.CallAfter(self._sudoku_frame.RecolourCell, colour, row_number,
                         column_number)
            self._sudoku_frame.GetLock().release()
        else:
            self.WriteIntoSocket('RSP', 'KO')

    def _handle_new_game(self, data):
        """
        Indicate whether the other player had the right to try
        to start a new game.
        """
        time = float(data)
        if self._is_peer_ask_ok(time):
            wx.CallAfter(self._sudoku_frame.DisplayQuestionMessage,
                         _('The other player wants to start a new game.') +
                         self._get_agree_question())
            if self._sudoku_frame.IsUserOk():
                self.WriteIntoSocket('RSP', 'OK')
                wx.CallAfter(self._sudoku_frame.SetEvtHandlerEnabled, False)
                wx.CallAfter(self._sudoku_frame.ResetInternalValues)
            else:
                self.WriteIntoSocket('RSP', 'KO')
        else:
            self.WriteIntoSocket('RSP', 'KO')

    def _handle_pause_game(self, data):
        """
        Indicate whether the other player had the right to try
        to pause the game.
        """
        time = float(data)
        if self._is_peer_ask_ok(time):
            self.WriteIntoSocket('RSP', 'OK')
            wx.CallAfter(self._sudoku_frame.PauseGame)
        else:
            self.WriteIntoSocket('RSP', 'KO')

    def _handle_allow_forcing_undo(self, data):
        """
        Indicate whether the other player had the right to try
        to undo an action done by the local player.
        """
        time = float(data)
        if self._is_peer_ask_ok(time):
            wx.CallAfter(self._sudoku_frame.DisplayQuestionMessage,
                         _('The other player wants to force an undo.') +
                         self._get_agree_question())
            if self._sudoku_frame.IsUserOk():
                self.WriteIntoSocket('RSP', 'OK')
            else:
                self.WriteIntoSocket('RSP', 'KO')
        else:
            self.WriteIntoSocket('RSP', 'KO')

    def _handle_undo(self):
        """Undo last action because the other player did it."""
        self._sudoku_frame.GetLock().acquire()
        self._sudoku_frame.ResetUndoNumber()
        wx.CallAfter(self._sudoku_frame.GetSudoku().GetHistory().MoveBack)
        self._sudoku_frame.GetLock().release()

    def _handle_redo(self):
        """Redo last undone action because the other player did it."""
        self._sudoku_frame.GetLock().acquire()
        wx.CallAfter(self._sudoku_frame.GetSudoku().GetHistory().MoveOn)
        self._sudoku_frame.ResetUndoNumber(False)
        self._sudoku_frame.GetLock().release()

    def _handle_retry(self, data):
        """
        Indicate whether the other player had the right to try
        to retry the game.
        """
        time = float(data)
        if self._is_peer_ask_ok(time):
            wx.CallAfter(self._sudoku_frame.DisplayQuestionMessage,
                         _('The other player wants to retry the game.') +
                         self._get_agree_question())
            if self._sudoku_frame.IsUserOk():
                self.WriteIntoSocket('RSP', 'OK')
                self._sudoku_frame.GetLock().acquire()
                wx.CallAfter(self._sudoku_frame.ResetInternalValues, False)
                wx.CallAfter(self._sudoku_frame.GetSudoku().GetHistory().Clear,
                             True)
                wx.CallAfter(self._sudoku_frame.ShowGridPanel, True)
                self._sudoku_frame.GetLock().release()
            else:
                self.WriteIntoSocket('RSP', 'KO')
        else:
            self.WriteIntoSocket('RSP', 'KO')

    def _handle_destroy_grid(self):
        """Destroy the grid because the other player did it."""
        wx.CallAfter(self._sudoku_frame.DestroyGrid, True)

    def _handle_create_grid(self, data):
        """Create a new grid because the other player did it."""
        size_of_box, size_of_side, x_sudoku = data.strip('[]').split(',')
        size_of_box, size_of_side = int(size_of_box), int(size_of_side)
        if x_sudoku == ' False':
            x_sudoku = False
        else:
            x_sudoku = True
        wx.CallAfter(self._sudoku_frame.SetFeaturesOfGame, size_of_box,
                     size_of_side, x_sudoku)
        wx.CallAfter(self._sudoku_frame.CreateGrid, True)
        wx.CallAfter(self._sudoku_frame.CreateNotEmptyGrid, True)

    def _handle_difficulty(self, data):
        """Set the difficulty of the sudoku puzzle."""
        wx.CallAfter(self._sudoku_frame.SetDifficulty, int(data))

    def _handle_end_of_creation(self):
        wx.CallAfter(self._sudoku_frame.GetSudoku().GetHistory().Clear)
        wx.CallAfter(self._sudoku_frame.EndOfCreationOfGrid, True)

    def _handle_response(self, data):
        """Handle peer response."""
        self._peer_response_lock.acquire()
        self._peer_ok = (data == 'OK')
        self._peer_response_lock.release()
        self._peer_responded.set()

    def _handle_value(self, data):
        """
        Set the value of a cell as the other player had the right
        to do it.
        """
        value, row_number, column_number = data.strip('[]').split(',')
        if value != 'None':
            value = int(value)
        else:
            value = None
        row_number, column_number = int(row_number), int(column_number)
        self._sudoku_frame.GetLock().acquire()
        wx.CallAfter(self._sudoku_frame.SetValue, value, row_number,
                     column_number, True)
        self._sudoku_frame.GetLock().release()

    def _handle_recolour(self, data):
        """Recolour a cell with the player's colour."""
        row_number, column_number = data.strip('[]').split(',')
        row_number, column_number = int(row_number), int(column_number)
        self._sudoku_frame.GetLock().acquire()
        colour = self._sudoku_frame.GetPlayerColour()
        wx.CallAfter(self._sudoku_frame.RecolourCell, colour, row_number,
                     column_number)
        self._sudoku_frame.GetLock().release()

    def _get_agree_question(self):
        """Get the question 'Do you agree?'"""
        return os.linesep + _('Do you agree?')

    def _is_peer_ask_ok(self, time):
        """Indicate whether the other player had the right to ask something."""
        self._sudoku_frame.GetLock().acquire()
        clock = self._sudoku_frame.GetClock()
        if clock is None:
            self._sudoku_frame.GetLock().release()
            return True
        else:
            local_delta = clock - self._your_time
            self._sudoku_frame.GetLock().release()
            peer_delta = time - self._other_player_time
            if local_delta < peer_delta or \
               (local_delta == peer_delta and self._sudoku_frame.IsServer()):
                return False
            else:
                return True

    def _end_of_connection(self):
        """React to an end of connection."""
        if not self._finished.isSet():
            self._sudoku_frame.NetworkProblemOccurred(_('Server stopped.'))


class SudokuServer(SudokuClient):
    """Sudoku server."""

    def __init__(self, sudoku_frame):
        SudokuClient.__init__(self, sudoku_frame, None)

    def stop(self):
        self._finished.set()
        if self._sock is not None:
            self._lock.acquire()
            if self._conn is None:
                self._stop_waiting_for_client()
            else:
                self._stop_waiting_for_data()
            self._close_master_socket()
            self._lock.release()
        self.join()

    def run(self):
        self._open_master_socket()
        if self._sock is not None:
            while not self._finished.isSet():
                SudokuClient.run(self)

    def IsThereAnyClient(self):
        """Indicate whether the server has a client."""
        self._lock.acquire()
        result = (self._conn is not None)
        self._lock.release()
        return result

    def _stop_waiting_for_client(self):
        """Stop waiting for a client."""
        if os.name == 'posix':
            self._sock.shutdown(socket.SHUT_RD)
        else:
            conn = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            try:
                conn.connect(('localhost', port))
                conn.send('QUIT')
            except socket.error:
                pass
            conn.close()

    def _stop_waiting_for_data(self):
        """Stop waiting for data."""
        SudokuClient._stop_waiting_for_data(self)
        if os.name == 'posix':
            self._sock.shutdown(socket.SHUT_RD)

    def _open_master_socket(self):
        """Open the socket."""
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            self._sock.bind((HOST, port))
            self._sock.listen(1)
            wx.CallAfter(self._sudoku_frame.DisplayMessage,
                         _('Server started.'), msg_colour)
        except socket.error:
            self._close_master_socket()
            self._sudoku_frame.NetworkProblemOccurred(
                                                    _('Cannot start server!'))

    def _close_master_socket(self):
        """Close the socket."""
        self._sock.close()
        self._sock = None

    def _open_socket(self):
        """Wait for a client."""
        try:
            self._conn, client = self._sock.accept()
            self._lock.acquire()
            if not self._finished.isSet():
                self._send_local_time()
                wx.CallAfter(self._sudoku_frame.ResetGrid)
                wx.CallAfter(self._sudoku_frame.DisplayMessage,
                             client[0] + _(' joined server.'), msg_colour)
        except socket.error:
            self._lock.acquire()
            self._conn = None
            if not self._finished.isSet():
                self._sudoku_frame.NetworkProblemOccurred(
                                                     _('Cannot join client!'))

    def _end_of_connection(self):
        """React to an end of connection."""
        wx.CallAfter(self._sudoku_frame.DisplayMessage,
                     _('Client disconnected.'), msg_colour)


class SudokuScrolledMessageDialog(wx.lib.dialogs.ScrolledMessageDialog):
    """Customised ScrolledMessageDialog used to display best scores."""

    def __init__(self, parent, msg, caption,
                 pos=wx.DefaultPosition, size=(500, 300),
                 style=wx.DEFAULT_DIALOG_STYLE):
        wx.Dialog.__init__(self, parent, -1, caption, pos, size, style)
        x, y = pos
        if x == -1 and y == -1:
            self.CenterOnScreen(wx.BOTH)
        self.text = text = wx.TextCtrl(self, -1, '',
                                       style=wx.TE_MULTILINE | wx.TE_READONLY)
        self._format_and_display(msg)
        ok = wx.Button(self, wx.ID_OK, 'OK')
        ok.SetDefault()
        lc = wx.lib.layoutf.Layoutf('t=t5#1;b=t5#2;l=l5#1;r=r5#1', (self, ok))
        text.SetConstraints(lc)
        lc = wx.lib.layoutf.Layoutf('b=b5#1;x%w50#1;w!80;h*', (self, ))
        ok.SetConstraints(lc)
        self.SetAutoLayout(1)
        self.Layout()

    @staticmethod
    def SetCustomStyle(text_ctrl, some_text, italic=True, bold=True,
                       colour=wx.NullColour):
        """Set a custom style."""
        tmppos = text_ctrl.GetInsertionPoint()
        text_ctrl.AppendText(some_text)
        tmppos2 = text_ctrl.GetInsertionPoint()
        font = wx.SystemSettings.GetFont(wx.SYS_SYSTEM_FONT)
        if bold:
            font.SetWeight(wx.BOLD)
            font.SetPointSize(8)
        else:
            font.SetPointSize(9)
        if italic:
            font.SetStyle(wx.ITALIC)
        text_ctrl.SetStyle(tmppos, tmppos2,
                           wx.TextAttr(colour, wx.NullColour, font))

    def _format_and_display(self, msg):
        """Format and display the message."""
        lines = msg.splitlines()
        position = 0
        for line in lines:
            if line.startswith('*****'):
                line = line.strip('*')
                position = self.text.GetInsertionPoint()
                SudokuScrolledMessageDialog.SetCustomStyle(self.text, line)
            else:
                SudokuScrolledMessageDialog.SetCustomStyle(self.text, line,
                                                           False, False)
            self.text.AppendText(os.linesep)
        self.text.SetInsertionPoint(position)


class SudokuIntValidator(wx.lib.intctrl.IntValidator):
    """Customised IntValidator that authorises tab traversal."""

    @classmethod
    def PreInit(cls, sudoku_frame):
        cls._sudoku_frame = sudoku_frame

    def OnChar(self, event):
        """Authorise navigation through the grid."""
        key = event.GetKeyCode()
        cell = event.GetEventObject()
        cell_id = cell.GetId()
        if key not in (wx.WXK_TAB, wx.WXK_UP, wx.WXK_DOWN,
                       wx.WXK_RIGHT, wx.WXK_LEFT):
            if self._sudoku_frame.Is2Players():
                self._sudoku_frame.GetLock().acquire()
            if cell.GetColors()[0] == self._sudoku_frame.GetPlayerColour():
                if self._sudoku_frame.Is2Players():
                    if cell.GetValue() is None:
                        if self._is_peer_ok(cell_id):
                            wx.lib.intctrl.IntValidator.OnChar(self, event)
                        else:
                            cell.SetColors(
                                      self._sudoku_frame.GetOtherPlayerColour(),
                                      self._sudoku_frame.GetOtherPlayerColour())
                    else:
                        wx.lib.intctrl.IntValidator.OnChar(self, event)
                    self._sudoku_frame.ResetClock()
                    self._sudoku_frame.GetLock().release()
                else:
                    wx.lib.intctrl.IntValidator.OnChar(self, event)
            elif self._sudoku_frame.Is2Players():
                self._sudoku_frame.GetLock().release()
        else:
            self._move_focus(key, cell_id)

    def _is_peer_ok(self, cell_id):
        """
        Wait for the response of the other player.

        Return the response.
        """
        self._sudoku_frame.SetClock()
        clock = self._sudoku_frame.GetClock()
        sudoku_peer = self._sudoku_frame.GetSudokuPeer()
        row_number = cell_id // self._sudoku_frame.GetSizeOfSide()
        column_number = cell_id % self._sudoku_frame.GetSizeOfSide()
        char_number = sudoku_peer.WriteIntoSocket('CHR', [clock, row_number,
                                                          column_number])
        if char_number is not None:
            self._sudoku_frame.GetLock().release()
            result = sudoku_peer.IsPeerOk()
            self._sudoku_frame.GetLock().acquire()
            return result
        else:
            return True

    def _move_focus(self, key, cell_id):
        """Put the focus on another cell."""
        if key == wx.WXK_TAB or key == wx.WXK_RIGHT:
            cell_id = (cell_id + 1) % self._sudoku_frame.GetSizeOfSide()**2
        elif key == wx.WXK_LEFT:
            cell_id = (cell_id - 1) % self._sudoku_frame.GetSizeOfSide()**2
        else:
            row_number = cell_id // self._sudoku_frame.GetSizeOfSide()
            column_number = cell_id % self._sudoku_frame.GetSizeOfSide()
            old_row_number = row_number
            if key == wx.WXK_UP:
                row_number = (row_number - 1) % \
                                             self._sudoku_frame.GetSizeOfSide()
                if old_row_number < row_number:
                    column_number = (column_number - 1) % \
                                             self._sudoku_frame.GetSizeOfSide()
            else:
                row_number = (row_number + 1) % \
                                             self._sudoku_frame.GetSizeOfSide()
                if old_row_number > row_number:
                    column_number = (column_number + 1) % \
                                             self._sudoku_frame.GetSizeOfSide()
            cell_id = row_number * self._sudoku_frame.GetSizeOfSide() + \
                                                                  column_number
        self._sudoku_frame.SetFocusedObject(
                                    self._sudoku_frame.FindWindowById(cell_id))


class SudokuFrame(wx.Frame):
    """The main frame of the application."""

    def __init__(self, path, window_title):
        SudokuIntValidator.PreInit(self)
        self._window_title = window_title
        self._init_pathes(path)
        self._get_saved_config()
        self._init_labels()
        self._init_puzzle()
        self._init_time()
        self._init_open_save()
        self._init_game_mode()
        self._init_gui()
        self._create_frame()
        self._sudoku = SudokuPuzzle(self, self._size_of_side, self._x_sudoku)
        self._assign_events()
        self._set_accelerators()
        if os.name == 'posix' and os.popen('pgrep pulseaudio').read() != '':
            pygame.mixer.init(8000, -16, 2, 2048)
        else:
            pygame.mixer.init(8000, -8, False, 1024 * 3)
        if os.name != 'nt':
            wx.MilliSleep(1000)

    def NetworkProblemOccurred(self, problem_msg):
        """Signal a network problem."""
        self._network_problem.set()
        self._network_problem_msg = problem_msg

    def SetValue(self, value, row_number, column_number, from_peer=False,
                 colour=None, to_peer=True):
        """Put a value at a cell."""
        cell_id = row_number * self._size_of_side + column_number
        cell = self.FindWindowById(cell_id)
        evt_handler_enabled = self.GetEvtHandlerEnabled()
        if evt_handler_enabled:
            self.SetEvtHandlerEnabled(False)
        if from_peer:
            self._add_to_history(value, row_number, column_number, cell)
        cell.SetValue(value)
        if self._generation:
            self._set_colour_when_generating(value, cell)
        if colour is not None:
            cell.SetColors(colour, colour)
        if evt_handler_enabled:
            self.SetEvtHandlerEnabled(True)
        if not self._generation:
            wx.YieldIfNeeded()
        if self._2_players and not from_peer and colour is None:
            if to_peer:
                self._sudoku_peer.WriteIntoSocket('VAL', [value, row_number,
                                                          column_number])
            else:
                self._sudoku_peer.WriteIntoSocket('CLR', [row_number,
                                                          column_number])
        if from_peer and not self._generation \
                     and self._sudoku.GetSudokuGrid().GetBlanksNumber() == 0:
            self._display_end_game()

    def GetSizeOfSide(self):
        """Get the size of a side of the Sudoku grid."""
        return self._size_of_side

    def SetFocusedObject(self, an_object):
        """Set the focus to an object."""
        self._focused_object = an_object

    def GetSendText(self):
        """Get the text control to send a message."""
        return self._send_text

    def OnSendMessage(self, event):
        """Send a message."""
        msg = self._send_text.GetValue()
        if msg != '':
            result = self._sudoku_peer.WriteIntoSocket('MSG', msg)
            if result is not None:
                self.DisplayMessage(msg, self._player_colour)
                self._send_text.Clear()

    def GetSudokuPeer(self):
        """Get the _sudoku_peer object."""
        return self._sudoku_peer

    def GetOtherPlayerColour(self):
        """Get the other player's colour."""
        if self._other_player_colour is None:
            if self._is_server:
                self._other_player_colour = client_colour
            else:
                self._other_player_colour = server_colour
        return self._other_player_colour

    def GetPlayerColour(self):
        """Get the player's colour."""
        return self._player_colour

    def Is2Players(self):
        """Indicate whether or not the game is in 2 players mode."""
        return self._2_players

    def IsServer(self):
        """Indicate whether or not the local game is the server."""
        return self._is_server

    def DisplayMessage(self, msg, colour):
        """Display messages in "_display_msg" text control."""
        SudokuScrolledMessageDialog.SetCustomStyle(self._display_msg,
                                                   time.strftime('%c') +
                                                   _(': ') + os.linesep, False,
                                                   colour=msg_colour)
        if colour != msg_colour:
            if colour == self._player_colour:
                player = self._you
            else:
                player = self._other_player
            SudokuScrolledMessageDialog.SetCustomStyle(self._display_msg,
                                                       player, False,
                                                       colour=colour)
        SudokuScrolledMessageDialog.SetCustomStyle(self._display_msg,
                                                   msg + os.linesep +
                                                   os.linesep, False, False,
                                                   colour)

    def GetLock(self):
        """Get the lock."""
        return self._lock

    def SetClock(self):
        """Set the clock."""
        self._clock = time.time()

    def ResetClock(self):
        """Reset the clock."""
        self._clock = None

    def GetClock(self):
        """Get the clock."""
        return self._clock

    def RecolourCell(self, colour, row_number, column_number):
        """Change the colour of the cell."""
        cell_id = row_number * self._size_of_side + column_number
        cell = self.FindWindowById(cell_id)
        self._old_colour = cell.GetColors()[0]
        cell.SetColors(colour, colour)

    def SetFeaturesOfGame(self, size_of_box, size_of_side, x_sudoku):
        """Set the '_size_of_box', the '_size_of_side' and '_x_sudoku'."""
        self._size_of_box = size_of_box
        self._set_sudoku_size(size_of_side)
        self._set_sudoku_type(x_sudoku)
        self._reset_sudoku()

    def SetDifficulty(self, difficulty):
        """Set the '_difficulty'."""
        self._difficulty = difficulty
        if self._difficulty == 0:
            self._difficulty_name = self._easy_label
        elif self._difficulty == 1:
            self._difficulty_name = self._medium_label
        elif self._difficulty == 2:
            self._difficulty_name = self._hard_label
        else:
            self._difficulty_name = self._very_hard_label
        self._set_game_title()

    def ShowGridPanel(self, show):
        """Show or hide the grid panel."""
        self._bkg_panel.SetSize(wx.Size(self._grid_panel.GetSize().GetWidth() +
                                        self._get_send_width(),
                                        self._grid_panel.GetSize().GetHeight()))
        self._grid_panel.Show(show)

    def ResetInternalValues(self, reset_player_colour=True):
        """Reset some attributes."""
        self.ShowGridPanel(False)
        self._cell = None
        self._undo_number = 0
        self._undo_number_history = []
        if reset_player_colour:
            if self._2_players:
                if self._is_server:
                    self._player_colour = server_colour
                else:
                    self._player_colour = client_colour
                self._other_player_colour = None
            else:
                self._player_colour = one_player_colour
        self._old_colour = self._player_colour
        if self._button_pause is not None:
            self._unpause_game()

    def DestroyGrid(self, from_peer=False):
        """Destroy the Sudoku grid."""
        self._count_time = False
        if self._2_players and not from_peer:
            self._sudoku_peer.WriteIntoSocket('DST', 'DESTROY')
        for i in range(self._size_of_side):
            for j in range(self._size_of_side):
                cell_id = i * self._size_of_side + j
                self.FindWindowById(cell_id).Destroy()
        for i in range(self._size_of_side):
            self._boxes[i].Destroy()

    def CreateGrid(self, from_peer=False):
        """Create the Sudoku grid."""
        if self._2_players and not from_peer:
            self._sudoku_peer.WriteIntoSocket('CRT', [self._size_of_box,
                                                      self._size_of_side,
                                                      self._x_sudoku])
        self._boxes = []
        self._resize()
        self._create_boxes()
        if not from_peer:
            self.SetEvtHandlerEnabled(False)
        self._create_cells()
        if not from_peer:
            self.SetEvtHandlerEnabled(True)
        self._boxes[0].SetFocus()

    def ResetGrid(self):
        """Reset the grid."""
        self.SetEvtHandlerEnabled(False)
        self.ResetInternalValues()
        self.DestroyGrid(True)
        self.CreateGrid(True)
        self._display_grid(False, False)
        self.SetEvtHandlerEnabled(True)

    def GetSudoku(self):
        """Get the '_sudoku'."""
        return self._sudoku

    def CreateNotEmptyGrid(self, from_peer=False):
        """
        Populates the Sudoku grid with some values
        if the user do not want an empty grid.
        """
        empty_grid = False
        if not from_peer:
            if not self._2_players:
                empty_grid = self._choice_for_empty_grid()
            if not empty_grid:
                self._choice_for_level_of_difficulty()
                if self._2_players:
                    self._sudoku_peer.WriteIntoSocket('DIF', self._difficulty)
            else:
                self.ShowGridPanel(True)
        if from_peer or not empty_grid:
            self._display_grid(False)
            msg = _('Please wait,') + os.linesep + \
                  _('generating the Sudoku puzzle...')
            self._label_generate_grid = wx.StaticText(self._grid_panel,
                                                      label=msg,
                                                      style=wx.ALIGN_CENTER)
            self._label_generate_grid.Centre()
            self._generation = True
            wx.YieldIfNeeded()
        if not from_peer:
            if not empty_grid:
                self.SetEvtHandlerEnabled(False)
                self._sudoku.CreateNewGame(self._difficulty)
            self.EndOfCreationOfGrid(empty_grid=empty_grid)

    def EndOfCreationOfGrid(self, from_peer=False, empty_grid=False):
        """Finish the creation of a non empty grid."""
        if self._2_players and not from_peer:
            self._sudoku_peer.WriteIntoSocket('EOC', 'END')
        if from_peer or not empty_grid:
            self.SetEvtHandlerEnabled(True)
            self._generation = False
            self._label_generate_grid.Destroy()
            self._display_grid(True)
        self._boxes[0].SetFocus()
        self._count_time = True
        self._begin_time = time.time()
        self._timer = None
        self._time_sep_cpt = 0

    def PauseGame(self):
        """Pause the game."""
        if self._boxes[0].IsShown():
            self._focused_component = self.FindFocus()
            self._count_time = False
            self._display_grid(False)
            self._button_pause = wx.Button(self._grid_panel, ID_PAUSE,
                                          label=_('PAUSE'))
            self._button_pause.Centre()
            self._button_pause.SetFocus()
            self._modify_bars(True)
        elif self._button_pause is not None:
            self._unpause_game()

    def DisplayQuestionMessage(self, msg):
        """Display a question."""
        question = wx.MessageDialog(self, msg, _('Question'),
                                    style=wx.YES_NO | wx.ICON_QUESTION)
        self._user_response = question.ShowModal()
        question.Destroy()
        self._user_responded.set()

    def IsUserOk(self):
        self._user_responded.wait()
        self._user_responded.clear()
        return (self._user_response == wx.ID_YES)

    def ResetUndoNumber(self, to_zero=True):
        """Reset the '_undo_number'."""
        if to_zero:
            self._undo_number_history.append(self._undo_number)
            self._undo_number = 0
        else:
            self._undo_number = self._undo_number_history.pop()

    def _init_pathes(self, path):
        """Initialize attributes related to the pathes."""
        self._user_path = os.path.expanduser('~')
        try:
            os.chdir(self._user_path)
        except:
            self._user_path = os.getcwd()
        self._path = path
        if os.name == 'nt':
            self._icon_path = os.path.join(self._path, 'icons', '12x12')
        else:
            self._icon_path = os.path.join(self._path, 'icons', '24x24')

    def _get_saved_config(self):
        """Get saved configuration."""
        self._user_cfg_dir = self._create_cfg_dir()
        self._user_cfg_file = os.path.join(self._user_cfg_dir, 'yapysudoku.cfg')
        self._config_parser = configparser.ConfigParser()
        if os.access(self._user_cfg_file, os.R_OK):
            user_cfg_file = open(self._user_cfg_file, 'r')
            self._config_parser.read_file(user_cfg_file)
            user_cfg_file.close()
            self._check_saved_cfg()
        else:
            self._create_cfg_game_section()
            self._create_cfg_snd_section()
            self._create_cfg_network_section()
            self._create_cfg_colours_section()
            self._create_cfg_forced_undo_section()

    def _set_cfg_game_size(self):
        """Set the default saved game size."""
        self._config_parser.set('Game', 'size', '9')

    def _set_cfg_game_type(self):
        """Set the default saved game type."""
        self._config_parser.set('Game', 'normal', 'True')

    def _set_cfg_game_difficulty(self):
        """Set the default saved game difficulty."""
        self._config_parser.set('Game', 'difficulty', '0')

    def _create_cfg_game_section(self):
        """Create the game section."""
        self._config_parser.add_section('Game')
        self._set_cfg_game_size()
        self._set_cfg_game_type()
        self._set_cfg_game_difficulty()

    def _set_cfg_sounds(self):
        """Set the default saved sound behaviour."""
        self._config_parser.set('Sounds', 'enabled', 'True')

    def _create_cfg_snd_section(self):
        """Create the sounds section."""
        self._config_parser.add_section('Sounds')
        self._set_cfg_sounds()

    def _set_cfg_network_port(self):
        """Set the default saved network port."""
        self._config_parser.set('Network', 'port', str(port))

    def _create_cfg_network_section(self):
        """Create the network section."""
        self._config_parser.add_section('Network')
        self._set_cfg_network_port()

    def _set_cfg_initially_filled_cell_colour(self):
        """Set the default saved 'initially_filled_cell' colour."""
        self._config_parser.set('Colours', 'initially_filled_cell',
                                str(initially_filled_cell_colour.GetRGB()))

    def _set_cfg_one_player_colour(self):
        """Set the default saved 'one_player' colour."""
        self._config_parser.set('Colours', 'one_player',
                                str(one_player_colour.GetRGB()))

    def _set_cfg_server_colour(self):
        """Set the default saved 'server' colour."""
        self._config_parser.set('Colours', 'server',
                                str(server_colour.GetRGB()))

    def _set_cfg_client_colour(self):
        """Set the default saved 'client' colour."""
        self._config_parser.set('Colours', 'client',
                                str(client_colour.GetRGB()))

    def _set_cfg_information_colour(self):
        """Set the default saved 'information' colour."""
        self._config_parser.set('Colours', 'information',
                                str(msg_colour.GetRGB()))

    def _set_cfg_diagonals_colour(self):
        """Set the default saved 'diagonals' colour."""
        self._config_parser.set('Colours', 'diagonals',
                                str(x_colour.GetRGB()))

    def _set_cfg_bg_colour(self):
        """Set the default saved 'background' colour."""
        self._config_parser.set('Colours', 'background',
                                str(bg_colour.GetRGB()))

    def _create_cfg_colours_section(self):
        """Create the colours section."""
        self._config_parser.add_section('Colours')
        self._create_cfg_colours_options()

    def _create_cfg_colours_options(self):
        """Create the colours options."""
        self._set_cfg_initially_filled_cell_colour()
        self._set_cfg_one_player_colour()
        self._set_cfg_server_colour()
        self._set_cfg_client_colour()
        self._set_cfg_information_colour()
        self._set_cfg_diagonals_colour()
        self._set_cfg_bg_colour()

    def _set_cfg_forced_undo(self):
        """Set the default 'forced undo' behaviour."""
        self._config_parser.set('Undo', 'forced', 'False')

    def _create_cfg_forced_undo_section(self):
        """Create the 'forced undo' section."""
        self._config_parser.add_section('Undo')
        self._set_cfg_forced_undo()

    def _check_saved_cfg(self):
        """
        Check saved config to correct it if some settings
        are missing or are incorrect.
        """
        cfg_sections = self._config_parser.sections()
        if 'Game' not in cfg_sections:
            self._create_cfg_game_section()
        else:
            self._check_cfg_game_section()
        if 'Sounds' not in cfg_sections:
            self._create_cfg_snd_section()
        else:
            self._check_cfg_snd_section()
        if 'Network' not in cfg_sections:
            self._create_cfg_network_section()
        else:
            self._check_cfg_network_section()
        if 'Colours' not in cfg_sections:
            self._create_cfg_colours_section()
        else:
            self._check_colours_section()
        if 'Undo' not in cfg_sections:
            self._create_cfg_forced_undo_section()
        else:
            self._check_forced_undo_section()

    def _check_cfg_game_section(self):
        """Check game section."""
        try:
            cfg_size = self._config_parser.getint('Game', 'size')
            if cfg_size != 4 and cfg_size != 9 and cfg_size != 16:
                self._set_cfg_game_size()
        except configparser.NoOptionError:
            self._set_cfg_game_size()
        except ValueError:
            self._set_cfg_game_size()
        try:
            self._config_parser.getboolean('Game', 'normal')
        except configparser.NoOptionError:
            self._set_cfg_game_type()
        except ValueError:
            self._set_cfg_game_type()
        try:
            cfg_difficulty = self._config_parser.getint('Game', 'difficulty')
            if cfg_difficulty < 0 or cfg_difficulty > 3:
                self._set_cfg_game_difficulty()
        except configparser.NoOptionError:
            self._set_cfg_game_difficulty()
        except ValueError:
            self._set_cfg_game_difficulty()

    def _check_cfg_snd_section(self):
        """Check sounds section."""
        try:
            self._config_parser.getboolean('Sounds', 'enabled')
        except configparser.NoOptionError:
            self._set_cfg_sounds()
        except ValueError:
            self._set_cfg_sounds()

    def _check_cfg_network_section(self):
        """Check network section."""
        try:
            cfg_network_port = self._config_parser.getint('Network', 'port')
            if cfg_network_port <= MIN_PORT or cfg_network_port >= MAX_PORT:
                self._set_cfg_network_port()
        except configparser.NoOptionError:
            self._set_cfg_network_port()
        except ValueError:
            self._set_cfg_network_port()

    def _check_colours_section(self):
        """Check colours section."""
        try:
            self._config_parser.getint('Colours', 'initially_filled_cell')
        except configparser.NoOptionError:
            self._set_cfg_initially_filled_cell_colour()
        except ValueError:
            self._set_cfg_initially_filled_cell_colour()
        try:
            self._config_parser.getint('Colours', 'one_player')
        except configparser.NoOptionError:
            self._set_cfg_one_player_colour()
        except ValueError:
            self._set_cfg_one_player_colour()
        try:
            self._config_parser.getint('Colours', 'server')
        except configparser.NoOptionError:
            self._set_cfg_server_colour()
        except ValueError:
            self._set_cfg_server_colour()
        try:
            self._config_parser.getint('Colours', 'client')
        except configparser.NoOptionError:
            self._set_cfg_client_colour()
        except ValueError:
            self._set_cfg_client_colour()
        try:
            self._config_parser.getint('Colours', 'information')
        except configparser.NoOptionError:
            self._set_cfg_information_colour()
        except ValueError:
            self._set_cfg_information_colour()
        try:
            self._config_parser.getint('Colours', 'diagonals')
        except configparser.NoOptionError:
            self._set_cfg_diagonals_colour()
        except ValueError:
            self._set_cfg_diagonals_colour()
        try:
            self._config_parser.getint('Colours', 'background')
        except configparser.NoOptionError:
            self._set_cfg_bg_colour()
        except ValueError:
            self._set_cfg_bg_colour()
        self._check_saved_colours()

    def _check_saved_colours(self):
        """Check colours options."""
        cfg_init_filled_cell = \
            self._config_parser.getint('Colours', 'initially_filled_cell')
        cfg_one_player = self._config_parser.getint('Colours', 'one_player')
        cfg_server = self._config_parser.getint('Colours', 'server')
        cfg_client = self._config_parser.getint('Colours', 'client')
        cfg_diag = self._config_parser.getint('Colours', 'diagonals')
        cfg_bg = self._config_parser.getint('Colours', 'background')
        if cfg_init_filled_cell == cfg_one_player or \
           cfg_init_filled_cell == cfg_server or \
           cfg_init_filled_cell == cfg_client or \
           cfg_init_filled_cell == cfg_diag or \
           cfg_init_filled_cell == cfg_bg or cfg_one_player == cfg_diag or \
           cfg_one_player == cfg_bg or cfg_server == cfg_client or \
           cfg_server == cfg_diag or cfg_server == cfg_bg or \
           cfg_client == cfg_diag or cfg_client == cfg_bg or cfg_diag == cfg_bg:
            self._create_cfg_colours_options()

    def _check_forced_undo_section(self):
        """Check sounds section."""
        try:
            self._config_parser.getboolean('Undo', 'forced')
        except configparser.NoOptionError:
            self._set_cfg_forced_undo()
        except ValueError:
            self._set_cfg_forced_undo()

    def _init_puzzle(self):
        """Initialize attributes related to the puzzle."""
        self._is_setting_title_possible = False
        self._set_sudoku_size(self._config_parser.getint('Game', 'size'))
        self._size_of_box = int(math.sqrt(self._size_of_side))
        self._set_sudoku_type(not self._config_parser.getboolean('Game', 'normal'))
        self.SetDifficulty(self._config_parser.getint('Game', 'difficulty'))
        self._generation = False

    def _init_time(self):
        """Initialize attributes related to the time."""
        self._begin_time = None
        self._duration = ['00', '00', '00']
        self._count_time = True
        self._time_separator = ':'
        self._time_sep_cpt = 0

    def _init_open_save(self):
        """Initialize attributes related to opening and saving a game."""
        self._file_ext = '.ysz'
        self._wildcard_ext = ' (*' + self._file_ext + ')|*' + self._file_ext
        self._prefix = {'x':'x_', 'h':'history_', 't':'time_',
                        'd':'difficulty_'}

    def _init_game_mode(self):
        """Initialize attributes related to the game mode."""
        global port
        port = self._config_parser.getint('Network', 'port')
        self._2_players = False
        self._player_colour = one_player_colour
        self._other_player_colour = None
        self._sudoku_peer = None
        self._lock = threading.RLock()
        self._clock = None
        self._timer = None
        self._user_responded = threading.Event()
        self._user_response = wx.ID_YES
        self._undo_number = 0
        self._undo_number_history = []
        self._old_colour = self._player_colour
        self._forced_undo = self._config_parser.getboolean('Undo', 'forced')
        self._focused_object = None
        self._cell = None
        self._allow_sounds = self._config_parser.getboolean('Sounds', 'enabled')
        self._button_pause = None
        self._label_generate_grid = None
        self._network_problem = threading.Event()
        self._network_problem_msg = None

    def _init_gui(self):
        """Initialize attributes related to the GUI."""
        global initially_filled_cell_colour, one_player_colour, \
               server_colour, client_colour, msg_colour, x_colour, bg_colour
        try:
            init_filled_cell_col = wx.Colour()
            init_filled_cell_col.SetRGB(
                self._config_parser.getint('Colours', 'initially_filled_cell'))
            one_player_col = wx.Colour()
            one_player_col.SetRGB(self._config_parser.getint('Colours',
                                                             'one_player'))
            server_col = wx.Colour()
            server_col.SetRGB(self._config_parser.getint('Colours', 'server'))
            client_col = wx.Colour()
            client_col.SetRGB(self._config_parser.getint('Colours', 'client'))
            msg_col = wx.Colour()
            msg_col.SetRGB(self._config_parser.getint('Colours', 'information'))
            x_col = wx.Colour()
            x_col.SetRGB(self._config_parser.getint('Colours', 'diagonals'))
            bg_col = wx.Colour()
            bg_col.SetRGB(self._config_parser.getint('Colours', 'background'))
            initially_filled_cell_colour = init_filled_cell_col
            one_player_colour = one_player_col
            server_colour = server_col
            client_colour = client_col
            msg_colour = msg_col
            x_colour = x_col
            bg_colour = bg_col
        except ValueError:
            self._create_cfg_colours_options()

    def _init_labels(self):
        """Initialize the labels."""
        self._init_menubar_labels()
        self._init_toolbar_labels()
        self._send_button_label = _('Send &message')
        self._error_label = _('Error')
        self._wildcard_label = _('YSZ Files')
        self._you_won = _('Congratulations, you won!')
        self._you = _('You: ')
        self._other_player = _('Other player: ')
        self._filled_cells = _(" filled cells.")
        self._normal_label = _('normal')
        self._with_diags_label = _('with diagonals')
        self._easy_label = _('easy')
        self._medium_label = _('medium')
        self._hard_label = _('hard')
        self._very_hard_label = _('very hard')

    def _init_menubar_labels(self):
        """Initialize the labels related to the menu bar."""
        self._init_game_menu_labels()
        self._init_opt_menu_labels()
        self._init_help_menu_labels()

    def _init_game_menu_labels(self):
        """Initialize the labels related to the game menu."""
        self._init_new_game_labels()
        self._init_open_save_labels()
        self._init_solve_labels()
        self._init_game_play_labels()
        self._best_menu_label = _('&Best scores')
        self._best_long_label = _('Display the best scores')
        self._best_ctrl_label = '\tCtrl+B'
        self._quit_menu_label = _('&Quit')
        self._quit_long_label = _('Quit this software')
        self._quit_ctrl_label = '\tCtrl+Q'
        self._game_menu_label = _('&Game')

    def _init_new_game_labels(self):
        """Initialize the labels related to a new game."""
        self._new_menu_label = _('&New')
        self._new_long_label = _('New game')
        self._new_ctrl_label = '\tCtrl+N'
        self._serv_menu_label = _('Start ser&ver')
        self._serv_long_label = _('Start server')
        self._stop_serv_menu_label = _('Stop ser&ver')
        self._stop_serv_long_label = _('Stop server')
        self._serv_ctrl_label = '\tCtrl+E'
        self._join_menu_label = _('&Join a server')
        self._join_long_label = _('Join a server')
        self._discon_menu_label = _('Dis&connect')
        self._discon_long_label = _('Disconnect from server')
        self._join_ctrl_label = '\tCtrl+J'

    def _init_open_save_labels(self):
        """Initialize the labels related to opening and saving a game."""
        self._open_menu_label = _('&Open')
        self._open_long_label = _('Open a YSZ file')
        self._open_ctrl_label = '\tCtrl+O'
        self._save_menu_label = _('&Save')
        self._save_long_label = _('Save into a YSZ file')
        self._save_ctrl_label = '\tCtrl+S'

    def _init_solve_labels(self):
        """Initialize the labels related to solving a game."""
        self._solve_menu_label = _('So&lve')
        self._solve_long_label = _('Solve the game')
        self._solve_ctrl_label = '\tCtrl+R'
        self._stop_menu_label = _('S&top resolution')
        self._stop_long_label = _('Stop resolution of the game')
        self._stop_ctrl_label = '\tCtrl+T'

    def _init_game_play_labels(self):
        """Initialize the labels related to the game play."""
        self._pause_menu_label = _('&Pause')
        self._pause_long_label = _('Pause the game')
        self._unpause_menu_label = _('Un&pause')
        self._unpause_long_label = _('Unpause the game')
        self._pause_ctrl_label = '\tCtrl+P'
        self._retry_menu_label = _('&Retry')
        self._retry_long_label = _('Retry this game')
        self._retry_ctrl_label = '\tCtrl+G'
        self._undo_menu_label = _('&Undo')
        self._undo_long_label = _('Undo action')
        self._undo_ctrl_label = '\tCtrl+Z'
        self._redo_menu_label = _('Re&do')
        self._redo_long_label = _('Redo action')
        self._redo_ctrl_label = '\tCtrl+Y'

    def _init_opt_menu_labels(self):
        """Initialize the labels related to the options menu."""
        self._sounds_menu_label = _('&Enable sounds')
        self._sounds_long_label = _('Enable/disable sounds')
        self._sounds_ctrl_label = '\tCtrl+M'
        self._port_menu_label = _('Change net&work port')
        self._port_long_label = _('Change the port for network game')
        self._port_ctrl_label = '\tCtrl+W'
        self._init_colour_submenu_labels()
        self._forced_undo_menu_label = _('Force undo')
        self._forced_undo_long_label = _('Try to force undo for network game')
        self._forced_undo_ctrl_label = '\tCtrl+C'
        self._options_menu_label = _('Opt&ions')

    def _init_colour_submenu_labels(self):
        """Initialize the labels related to the colours submenu."""
        self._colours_menu_label = _('Change di&fferent colours')
        self._colours_long_label = _('Change different colours')
        self._init_filled_colour_menu_label = \
                                            _('Initiall&y filled cell colour')
        self._init_filled_colour_long_label = \
                        _('Colour for filled cell at the beginning of a game')
        self._init_filled_colour_ctrl_label = '\tCtrl+F'
        self._x_colour_menu_label = _('Colour in &X mode')
        self._x_colour_long_label = \
                                _('Background colour for diagonals in X mode')
        self._x_colour_ctrl_label = '\tCtrl+X'
        self._bg_colour_menu_label = _('Default bac&kground colour')
        self._bg_colour_long_label = _('Default background colour for cells')
        self._bg_colour_ctrl_label = '\tCtrl+D'
        self._one_player_colour_menu_label = _('Colour in one-player game')
        self._one_player_colour_long_label = \
                                        _('Player\'s colour in one-player game')
        self._one_player_colour_ctrl_label = '\tCtrl+0'
        self._server_colour_menu_label = _('Server colour in network game')
        self._server_colour_long_label = \
                            _('Server side player\'s colour in network game')
        self._server_colour_ctrl_label = '\tCtrl+1'
        self._client_colour_menu_label = _('Client colour in network game')
        self._client_colour_long_label = \
                            _('Client side player\'s colour in network game')
        self._client_colour_ctrl_label = '\tCtrl+2'
        self._msg_colour_menu_label = _('Text colour in network game')
        self._msg_colour_long_label = \
                        _('Colour for information displayed in network game')
        self._msg_colour_ctrl_label = '\tCtrl+I'
        self._reset_colours_menu_label = _('Reset all colours')
        self._reset_colours_long_label = \
                                    _('Reset colours to their default values')
        self._reset_colours_ctrl_label = '\tCtrl+V'

    def _init_help_menu_labels(self):
        """Initialize the labels related to the help menu."""
        self._about_menu_label = _('&About')
        self._about_long_label = _('About this software')
        self._about_ctrl_label = '\tCtrl+A'
        if wx.Platform != '__WXGTK__':
            self._license_menu_label = _('License')
            self._license_long_label = _('Software license')
            self._license_ctrl_label = '\tCtrl+L'
        self._help_menu_label = _('&Help')

    def _init_toolbar_labels(self):
        """Initialize the labels related to the tool bar."""
        self._new_short_label = _('New')
        self._open_short_label = _('Open')
        self._save_short_label = _('Save')
        self._pause_short_label = _('Pause')
        self._unpause_short_label = _('Unpause')
        self._solve_short_label = _('Solve')
        self._stop_short_label = _('Stop resolution')
        self._retry_short_label = _('Retry')
        self._undo_short_label = _('Undo')
        self._redo_short_label = _('Redo')
        self._about_short_label = _('About')
        self._quit_short_label = _('Quit')
        self._discon_short_label = _('Disconnect')

    def _add_to_history(self, value, row_number, column_number, cell):
        """Add to history because the other player did it."""
        old_value = cell.GetValue()
        if value is None:
            new_colour = self._player_colour
        elif self._generation:
            new_colour = initially_filled_cell_colour
        else:
            new_colour = self.GetOtherPlayerColour()
        cell.SetColors(new_colour, new_colour)
        self._sudoku.GetHistory().Push(row_number, column_number,
                                       old_value, value, False,
                                       self._old_colour, new_colour)
        self._old_colour = new_colour

    def _reset_sudoku(self):
        """Set a new SudokuPuzzle to '_sudoku'."""
        self._sudoku = SudokuPuzzle(self, self._size_of_side, self._x_sudoku)

    def _set_colour_when_generating(self, value, cell):
        """Set the colour of the cell when generating a new grid."""
        if value is None:
            cell.SetColors(self._player_colour, self._player_colour)
        else:
            cell.SetColors(initially_filled_cell_colour,
                           initially_filled_cell_colour)

    def _unpause_game(self):
        """Unpause the game."""
        self._button_pause.Destroy()
        self._modify_bars(False)
        self._display_grid(True)
        if self._focused_component is not None:
            self._focused_component.SetFocus()
        self._button_pause = None
        self._set_time(self._duration)
        self._count_time = True

    def _choice_for_size_of_sudoku(self):
        """Open a pop up to choose the size of the sudoku."""
        choice = wx.SingleChoiceDialog(None,
                                       _('Choose size of Sudoku puzzle:'),
                                       _('Size of Sudoku puzzle'),
                                       ['4', '9', '16'])
        choice.SetSelection(self._size_of_box - 2)
        if os.name != 'nt':
            choice.SetSize(wx.Size(choice.GetSize().GetWidth(), 250))
        result = choice.ShowModal()
        if result == wx.ID_OK:
            self._set_sudoku_size(int(choice.GetStringSelection()))
            self._size_of_box = int(math.sqrt(self._size_of_side))
        choice.Destroy()

    def _choice_for_type_of_sudoku(self):
        """Open a pop up to choose the type of the sudoku."""
        choice = wx.SingleChoiceDialog(None,
                                       _('Choose type of Sudoku puzzle:'),
                                       _('Type of Sudoku puzzle'),
                                       [self._normal_label,
                                        self._with_diags_label])
        if self._x_sudoku:
            choice.SetSelection(1)
        else:
            choice.SetSelection(0)
        if os.name != 'nt':
            choice.SetSize(wx.Size(choice.GetSize().GetWidth(), 250))
        result = choice.ShowModal()
        if result == wx.ID_OK:
            self._set_sudoku_type(
                            choice.GetStringSelection() != self._normal_label)
        choice.Destroy()

    def _choice_for_empty_grid(self):
        """Open a pop up to choose if the grid is empty or not."""
        empty_grid = True
        yes = _('yes')
        choice = wx.SingleChoiceDialog(None, _('Do you want an empty grid?'),
                                       _('Empty grid'), [_('no'), yes])
        if os.name != 'nt':
            choice.SetSize(wx.Size(choice.GetSize().GetWidth(), 250))
        result = choice.ShowModal()
        if result == wx.ID_OK:
            empty_grid = (choice.GetStringSelection() == yes)
        choice.Destroy()
        return empty_grid

    def _choice_for_level_of_difficulty(self):
        """Open a pop up to choose if the level of difficulty."""
        choice = wx.SingleChoiceDialog(None, _('Choose difficulty level:'),
                                       _('Difficulty level'),
                                       [self._easy_label, self._medium_label,
                                        self._hard_label,
                                        self._very_hard_label])
        choice.SetSelection(self._difficulty)
        if os.name != 'nt':
            choice.SetSize(wx.Size(choice.GetSize().GetWidth(), 250))
        result = choice.ShowModal()
        if result == wx.ID_OK:
            self.SetDifficulty(choice.GetSelection())
        choice.Destroy()

    def _set_sudoku_size(self, size_of_side):
        """Set the size of the sudoku."""
        self._size_of_side = size_of_side
        self._set_game_title()

    def _set_sudoku_type(self, x_sudoku):
        """Set the type of sudoku."""
        self._x_sudoku = x_sudoku
        if self._x_sudoku:
            self._x_sudoku_name = self._with_diags_label
        else:
            self._x_sudoku_name = self._normal_label
        self._set_game_title()

    def _set_game_title(self):
        """Set the title of the game."""
        if self._is_setting_title_possible:
            game_title = self._window_title + ' (' + str(self._size_of_side) + \
                         'x'  + str(self._size_of_side) + ', ' + \
                         self._x_sudoku_name + ', ' + \
                         self._difficulty_name + ')'
            self.SetTitle(game_title)

    def _create_cfg_dir(self):
        """
        Create the directory containing the config files.

        Return the directory containing the config files.
        """
        user_cfg_dir = os.path.join(self._user_path, '.yapysudoku')
        if not os.access(user_cfg_dir, os.F_OK):
            os.mkdir(user_cfg_dir)
        return user_cfg_dir

    def _create_frame(self):
        """Create the frame."""
        wx.Frame.__init__(self, None,
                          style=wx.MINIMIZE_BOX | wx.MAXIMIZE_BOX |
                                wx.SYSTEM_MENU | wx.CAPTION | wx.CLOSE_BOX |
                                wx.CLIP_CHILDREN | wx.RESIZE_BORDER)
        self._is_setting_title_possible = True
        self._set_game_title()
        if os.name == 'nt':
            ib = wx.IconBundle()
            ib.AddIconFromFile(os.path.join(self._path, 'icons',
                                            'yapysudoku.ico'),
                               wx.BITMAP_TYPE_ICO)
            self.SetIcons(ib)
        else:
            self.SetIcon(wx.Icon(os.path.join(self._path, 'icons',
                                              'yapysudoku.png'),
                                 wx.BITMAP_TYPE_PNG))
        self.Centre()
        self._create_menu_bar()
        self._create_tool_bar()
        self._create_status_bar()
        self._create_background_panel()
        self._create_send_panel()
        self._create_grid_panel()
        self.Show(True)

    def _create_background_panel(self):
        """Create the background panel."""
        self._bkg_panel = wx.Panel(self)
        image_file = os.path.join(self._path, 'main_bkg.png')
        if os.path.exists(image_file):
            bmp = wx.Bitmap(image_file)
            self._bkg_panel.bitmap = wx.StaticBitmap(self._bkg_panel,
                                                     wx.ID_ANY, bmp, (0, 0))

    def _create_menu_bar(self):
        """Create the menu bar."""
        game_menu = self._create_game_menu()
        options_menu = self._create_options_menu()
        info_menu = self._create_info_menu()
        menus = wx.MenuBar()
        menus.Append(game_menu, self._game_menu_label)
        menus.Append(options_menu, self._options_menu_label)
        menus.Append(info_menu, self._help_menu_label)
        self.SetMenuBar(menus)

    def _create_game_menu(self):
        """
        Create the game menu.

        Return this menu.
        """
        game_menu = wx.Menu()
        game_menu.Append(wx.ID_NEW,
                         self._new_menu_label + self._new_ctrl_label,
                         self._new_long_label)
        game_menu.Append(ID_SERVER,
                         self._serv_menu_label + self._serv_ctrl_label,
                         self._serv_long_label)
        game_menu.Append(ID_JOIN,
                         self._join_menu_label + self._join_ctrl_label,
                         self._join_long_label)
        game_menu.AppendSeparator()
        game_menu.Append(wx.ID_OPEN,
                         self._open_menu_label + self._open_ctrl_label,
                         self._open_long_label)
        game_menu.Append(wx.ID_SAVEAS,
                         self._save_menu_label + self._save_ctrl_label,
                         self._save_long_label)
        self._menu_separators = []
        self._menu_separators.append(game_menu.AppendSeparator())
        game_menu.Append(wx.ID_HIGHEST,
                         self._best_menu_label + self._best_ctrl_label,
                         self._best_long_label)
        self._menu_separators.append(game_menu.AppendSeparator())
        game_menu.Append(ID_PAUSE,
                         self._pause_menu_label + self._pause_ctrl_label,
                         self._pause_long_label)
        game_menu.AppendSeparator()
        game_menu.Append(ID_STARTUP,
                         self._solve_menu_label + self._solve_ctrl_label,
                         self._solve_long_label)
        game_menu.Append(wx.ID_STOP,
                         self._stop_menu_label + self._stop_ctrl_label,
                         self._stop_long_label)
        self._menu_separators.append(game_menu.AppendSeparator())
        game_menu.Append(wx.ID_RETRY,
                         self._retry_menu_label + self._retry_ctrl_label,
                         self._retry_long_label)
        game_menu.Append(wx.ID_UNDO,
                         self._undo_menu_label + self._undo_ctrl_label,
                         self._undo_long_label)
        game_menu.Append(wx.ID_REDO,
                         self._redo_menu_label + self._redo_ctrl_label,
                         self._redo_long_label)
        game_menu.AppendSeparator()
        game_menu.Append(wx.ID_EXIT,
                         self._quit_menu_label + self._quit_ctrl_label,
                         self._quit_long_label)
        return game_menu

    def _create_options_menu(self):
        """
        Create the options menu.

        Return this menu.
        """
        options_menu = wx.Menu()
        sounds_menuitem = options_menu.AppendCheckItem(wx.ID_PREFERENCES,
                                                       self._sounds_menu_label +
                                                       self._sounds_ctrl_label,
                                                       self._sounds_long_label)
        if self._allow_sounds:
            sounds_menuitem.Check()
        options_menu.AppendSeparator()
        options_menu.Append(ID_PORT,
                            self._port_menu_label + self._port_ctrl_label,
                            self._port_long_label)
        options_menu.AppendSeparator()
        colours_menu = self._create_colours_submenu()
        options_menu.AppendSubMenu(colours_menu, self._colours_menu_label,
                                   self._colours_long_label)
        options_menu.AppendSeparator()
        forced_undo_menuitem = \
            options_menu.AppendCheckItem(ID_FORCED_UNDO,
                                         self._forced_undo_menu_label +
                                         self._forced_undo_ctrl_label,
                                         self._forced_undo_long_label)
        if self._forced_undo:
            forced_undo_menuitem.Check()
        return options_menu

    def _create_colours_submenu(self):
        """
        Create the colours submenu.

        Return this menu.
        """
        colours_menu = wx.Menu()
        colours_menu.Append(ID_BG_COLOUR, self._bg_colour_menu_label +
                            self._bg_colour_ctrl_label,
                            self._bg_colour_long_label)
        colours_menu.Append(ID_X_COLOUR, self._x_colour_menu_label +
                            self._x_colour_ctrl_label,
                            self._x_colour_long_label)
        colours_menu.AppendSeparator()
        colours_menu.Append(ID_INIT_COLOUR,
                            self._init_filled_colour_menu_label +
                            self._init_filled_colour_ctrl_label,
                            self._init_filled_colour_long_label)
        colours_menu.AppendSeparator()
        colours_menu.Append(ID_ONE_P_COLOUR,
                            self._one_player_colour_menu_label +
                            self._one_player_colour_ctrl_label,
                            self._one_player_colour_long_label)
        colours_menu.AppendSeparator()
        colours_menu.Append(ID_SERV_P_COLOUR, self._server_colour_menu_label +
                            self._server_colour_ctrl_label,
                            self._server_colour_long_label)
        colours_menu.Append(ID_CLI_P_COLOUR, self._client_colour_menu_label +
                            self._client_colour_ctrl_label,
                            self._client_colour_long_label)
        colours_menu.AppendSeparator()
        colours_menu.Append(ID_MSG_COLOUR, self._msg_colour_menu_label +
                            self._msg_colour_ctrl_label,
                            self._msg_colour_long_label)
        colours_menu.AppendSeparator()
        colours_menu.Append(ID_RST_COLOURS, self._reset_colours_menu_label +
                            self._reset_colours_ctrl_label,
                            self._reset_colours_long_label)
        return colours_menu

    def _create_info_menu(self):
        """
        Create the info menu.

        Return this menu.
        """
        info_menu = wx.Menu()
        info_menu.Append(wx.ID_ABOUT,
                         self._about_menu_label + self._about_ctrl_label,
                         self._about_long_label)
        if wx.Platform != '__WXGTK__':
            info_menu.Append(ID_LICENSE, self._license_menu_label +
                             self._license_ctrl_label,
                             self._license_long_label)
        return info_menu

    def _create_bitmap(self, icon_name=None):
        """Create a bitmap."""
        if wx.__version__ > '4.1.1':
            if icon_name is None:
                return wx.BitmapBundle(wx.NullBitmap)
            else:
                return wx.BitmapBundle(wx.Bitmap(os.path.join(self._icon_path,
                                                              icon_name)))
        else:
            if icon_name is None:
                return wx.NullBitmap
            else:
                return wx.Bitmap(os.path.join(self._icon_path, icon_name))

    def _create_tool_bar(self):
        """Create the tool bar."""
        buttons = wx.ToolBar(self)
        buttons.AddTool(wx.ID_NEW, self._new_short_label,
                        self._create_bitmap('new.png'), self._create_bitmap(),
                        shortHelp=self._new_short_label,
                        longHelp=self._new_long_label)
        buttons.AddTool(wx.ID_OPEN, self._open_short_label,
                        self._create_bitmap('open.png'), self._create_bitmap(),
                        shortHelp=self._open_short_label,
                        longHelp=self._open_long_label)
        buttons.AddTool(wx.ID_SAVEAS, self._save_short_label,
                        self._create_bitmap('save_as.png'),
                        self._create_bitmap(),
                        shortHelp=self._save_short_label,
                        longHelp=self._save_long_label)
        buttons.AddSeparator()
        buttons.AddTool(ID_PAUSE, self._pause_short_label,
                        self._create_bitmap('pause.png'), self._create_bitmap(),
                        shortHelp=self._pause_short_label,
                        longHelp=self._pause_long_label)
        buttons.AddSeparator()
        buttons.AddTool(ID_STARTUP, self._solve_short_label,
                        self._create_bitmap('solve.png'), self._create_bitmap(),
                        shortHelp=self._solve_short_label,
                        longHelp=self._solve_long_label)
        buttons.AddTool(wx.ID_STOP, self._stop_short_label,
                        self._create_bitmap('stop.png'), self._create_bitmap(),
                        shortHelp=self._stop_short_label,
                        longHelp=self._stop_long_label)
        buttons.AddSeparator()
        buttons.AddTool(wx.ID_RETRY, self._retry_short_label,
                        self._create_bitmap('retry.png'), self._create_bitmap(),
                        shortHelp=self._retry_short_label,
                        longHelp=self._retry_long_label)
        buttons.AddTool(wx.ID_UNDO, self._undo_short_label,
                        self._create_bitmap('undo.png'), self._create_bitmap(),
                        shortHelp=self._undo_short_label,
                        longHelp=self._undo_long_label)
        buttons.AddTool(wx.ID_REDO, self._redo_short_label,
                        self._create_bitmap('redo.png'), self._create_bitmap(),
                        shortHelp=self._redo_short_label,
                        longHelp=self._redo_long_label)
        buttons.AddSeparator()
        buttons.AddTool(wx.ID_ABOUT, self._about_short_label,
                        self._create_bitmap('help.png'), self._create_bitmap(),
                        shortHelp=self._about_short_label,
                        longHelp=self._about_long_label)
        buttons.AddTool(wx.ID_EXIT, self._quit_short_label,
                        self._create_bitmap('exit.png'), self._create_bitmap(),
                        shortHelp=self._quit_short_label,
                        longHelp=self._quit_long_label)
        buttons.Realize()
        self.SetToolBar(buttons)

    def _create_status_bar(self):
        """Create the status bar."""
        self.CreateStatusBar(2, id=-1)
        self.SetStatusWidths([-1, 120])

    def _get_size_of_cell(self):
        """Return the size of a cell."""
        if os.name == 'nt':
            return 40
        else:
            return 46

    def _get_grid_width(self):
        """Return the width of the grid."""
        if self._2_players:
            if os.name == 'nt':
                return self._size_of_side * self._get_size_of_cell() + \
                       self._size_of_box * 4 + 20
            else:
                return self._size_of_side * self._get_size_of_cell() + \
                       self._size_of_box * 4 + 46
        elif os.name == 'nt':
            return max(self._size_of_side * self._get_size_of_cell() +
                       self._size_of_box * 4 + 22, 382)
        else:
            return max(self._size_of_side * self._get_size_of_cell() +
                       self._size_of_box * 4 + 16, 474)

    def _get_grid_height(self):
        """Return the height of the grid."""
        if os.name == 'nt':
            return self._size_of_side * self._get_size_of_cell() + \
                   self._size_of_box * 4 + 20
        else:
            return self._size_of_side * self._get_size_of_cell() + \
                   self._size_of_box * 4 + 44

    def _create_grid_panel(self):
        """Create the panel that contains the grid."""
        self._grid_panel = wx.Panel(self._bkg_panel,
                                    size=(self._get_grid_width(),
                                          self._get_grid_height()),
                                    pos=(0, 0))
        self.CreateGrid()
        self._display_grid(False, False)

    def _get_send_width(self):
        """Return the width of the the panel to send messages."""
        return 350

    def _create_send_panel(self):
        """Create the panel to send messages."""
        self._send_panel = wx.Panel(self._bkg_panel,
                                    size=(self._get_send_width(),
                                          self._get_grid_height()),
                                    pos=(self._get_grid_width(), 0))
        if not self._2_players:
            self._send_panel.Show(False)
        self._create_send_text()
        self._create_send_button()
        self._create_display_msg()

    def _create_send_text(self):
        """Create the text control to send messages."""
        if os.name == 'nt':
            delta = 18
            hpos = 4
        else:
            delta = 4
            hpos = 10
        self._send_text = wx.TextCtrl(self._send_panel,
                                      size=(self._get_send_width() - delta, 20),
                                      pos=(0, hpos), style=wx.TE_PROCESS_ENTER)

    def _create_send_button(self):
        """Create the button to send messages."""
        if os.name == 'nt':
            hpos = 28
        else:
            hpos = 36
        wx.Button(self._send_panel, id=ID_SEND, label=self._send_button_label,
                  pos=(100, hpos))

    def _get_wdelta(self):
        """Return the delta for width."""
        if os.name == 'nt':
            return 18
        else:
            return 4

    def _get_hdelta(self):
        """Return the delta for height."""
        if os.name == 'nt':
            return 72
        else:
            return 110

    def _create_display_msg(self):
        """Create the text control to display messages."""
        if os.name == 'nt':
            hpos = 56
        else:
            hpos = 74
        self._display_msg = wx.TextCtrl(self._send_panel,
                                        size=(self._get_send_width() - \
                                              self._get_wdelta(),
                                              self._get_grid_height() - \
                                              self._get_hdelta()),
                                        pos=(0, hpos),
                                        style=wx.TE_MULTILINE | wx.TE_READONLY)

    def _display_grid(self, display, display_grid_panel=True):
        """Do not display the grid."""
        self.ShowGridPanel(display_grid_panel)
        for i in range(self._size_of_side):
            self._boxes[i].Show(display)

    def _create_boxes(self):
        """Create the Sudoku boxes."""
        for i in range(self._size_of_box):
            for j in range(self._size_of_box):
                position = None
                if os.name == 'nt':
                    position = (j * (self._get_size_of_cell() * \
                                     self._size_of_box + 5),
                                i * (self._get_size_of_cell() * \
                                     self._size_of_box + 5))
                elif self._size_of_box == 3 and not self._2_players:
                    position = (j * (self._get_size_of_cell() * \
                                     self._size_of_box + 5) + 20,
                                i * (self._get_size_of_cell() * \
                                     self._size_of_box + 5) + 4)
                else:
                    position = (j * (self._get_size_of_cell() * \
                                     self._size_of_box + 5) + 4,
                                i * (self._get_size_of_cell() * \
                                     self._size_of_box + 5) + 4)
                if os.name == 'nt':
                    delta = 4
                else:
                    delta = 8
                self._boxes.append(wx.Panel(self._grid_panel,
                                            style=wx.SUNKEN_BORDER,
                                            size=(self._get_size_of_cell() * \
                                                  self._size_of_box + delta,
                                                  self._get_size_of_cell() * \
                                                  self._size_of_box + delta),
                                            pos=position))

    def _create_cells(self):
        """Create the Sudoku cells."""
        for i in range(self._size_of_side):
            for j in range(self._size_of_side):
                x_coord = j % self._size_of_box
                y_coord = i % self._size_of_box
                box_number = (i // self._size_of_box) * self._size_of_box + \
                             j // self._size_of_box
                cell = wx.lib.intctrl.IntCtrl(self._boxes[box_number],
                                              i * self._size_of_side + j,
                                              size=(self._get_size_of_cell(), \
                                                    self._get_size_of_cell()),
                                              pos=(x_coord * \
                                                   self._get_size_of_cell(),
                                                   y_coord * \
                                                   self._get_size_of_cell()),
                                              style=wx.SUNKEN_BORDER |
                                                    wx.ALIGN_CENTER_HORIZONTAL,
                                              value=None, min=1,
                                              max=self._size_of_side,
                                              limited=True, allow_none=True,
                                              validator=SudokuIntValidator())
                cell.SetFont(wx.Font(18, wx.SWISS, wx.NORMAL, wx.BOLD))
                cell.SetColors(self._player_colour, self._player_colour)
                if self._x_sudoku and \
                   (i == j or i + j + 1 == self._size_of_side):
                    cell.SetBackgroundColour(x_colour)
                else:
                    cell.SetBackgroundColour(bg_colour)

    def _get_frame_width(self):
        """Return the width of the frame."""
        if self._2_players:
            return self._get_grid_width() + self._get_send_width()
        else:
            return self._get_grid_width()

    def _get_frame_height(self):
        """Return the height of the frame."""
        if os.name != 'nt':
            return self._get_grid_height() + 88
        else:
            return self._get_grid_height() + 96

    def _resize(self):
        """Resize the Sudoku frame."""
        self._grid_panel.SetSize(wx.Size(self._get_grid_width(),
                                         self._get_grid_height()))
        if self._size_of_box == 3 and os.name != 'nt':
            if self._2_players:
                delta = -16
            else:
                delta = 16
            for box in self._boxes:
                box.SetPosition(wx.Point(box.GetPosition().Get()[0] + delta,
                                         box.GetPosition().Get()[1]))
        if os.name != 'nt':
            self._send_panel.SetSize(wx.Size(self._get_send_width(),
                                             self._get_grid_height()))
        else:
            self._send_panel.SetSize(wx.Size(self._get_send_width(),
                                             self._get_grid_height() - 12))
        self._send_panel.SetPosition(wx.Point(self._get_grid_height(), 0))
        self._display_msg.SetSize(wx.Size(self._get_send_width() - \
                                          self._get_wdelta(),
                                          self._get_grid_height() - \
                                          self._get_hdelta()))
        self.SetSize(wx.Size(self._get_frame_width(), self._get_frame_height()))
        self.Centre()
        if self._button_pause is not None:
            self._button_pause.Centre()

    def _set_duration(self, count_time=True):
        """Set the duration of the game."""
        if count_time:
            if self._count_time:
                self._duration = \
                         list(time.gmtime(time.time() - self._begin_time)[3:6])
                for i in range(3):
                    self._duration[i] = str(self._duration[i]).rjust(2, '0')
        else:
            self._count_time = count_time

    def _duration_to_string(self, duration, time_separator=None):
        """Return a string representation of a duration."""
        if time_separator is not None:
            return duration[0] + time_separator + duration[1] + \
                   time_separator + duration[2]
        else:
            return duration[0] + self._time_separator + duration[1] + \
                   self._time_separator + duration[2]

    def _display_duration(self):
        """Display duration of the game."""
        if self._begin_time is None:
            duration = '00:00:00'
        else:
            self._set_duration()
            duration = self._duration_to_string(self._duration)
            if self._count_time:
                self._time_sep_cpt += 1
                if self._time_sep_cpt == 50:
                    self._time_sep_cpt = 0
                    if self._time_separator == ':':
                        self._time_separator = ' '
                    else:
                        self._time_separator = ':'
            else:
                self._time_separator = ':'
        self.SetStatusText(duration, 1)

    def _set_size_of_sudoku(self, path):
        """Set the size of the sudoku."""
        csv_file = open(path, 'r', newline='')
        csv_reader = csv.reader(csv_file)
        for row in csv_reader:
            self._set_sudoku_size(len(row))
            break
        csv_file.close()
        self._size_of_box = int(math.sqrt(self._size_of_side))
        self._config_parser.set('Game', 'size', str(self._size_of_side))

    def _set_type_of_sudoku(self, path):
        """Set the type of the sudoku."""
        a_file = open(path, 'r')
        self._set_sudoku_type(int(a_file.readline()) == 1)
        a_file.close()
        self._config_parser.set('Game', 'normal', str(not self._x_sudoku))

    def _set_history_of_sudoku(self, path):
        """
        Set the history of the sudoku.

        Return the trace of the SudokuHistory object
        and the position in the history.
        """
        a_file = open(path, 'r', newline='')
        csv_reader = csv.reader(a_file)
        first_line = True
        position = -1
        trace = []
        for row in csv_reader:
            if first_line:
                first_line = False
                position = int(row[0])
            else:
                old_value = row[2]
                if old_value == '':
                    old_value = None
                else:
                    old_value = int(old_value)
                new_value = row[3]
                if new_value == '':
                    new_value = None
                else:
                    new_value = int(new_value)
                trace.append(SudokuHistoryElement(int(row[0]), int(row[1]),
                                                  old_value, new_value,
                                                  self._player_colour,
                                                  self._player_colour))
        a_file.close()
        return [position, trace]

    def _set_duration_of_sudoku(self, path):
        """Set the duration of the sudoku."""
        a_file = open(path, 'r', newline='')
        csv_reader = csv.reader(a_file)
        for row in csv_reader:
            self._set_time(row)
        a_file.close()

    def _set_difficulty_of_sudoku(self, path):
        """Set the difficulty level of the sudoku."""
        a_file = open(path, 'r')
        self.SetDifficulty(int(a_file.readline()))
        a_file.close()
        self._config_parser.set('Game', 'difficulty', str(self._difficulty))

    def _set_time(self, duration):
        """Set the timer value."""
        self._count_time = False
        if duration[0] != '':
            cur_time = time.time()
            self._begin_time = cur_time - int(duration[0]) * 3600 - \
                               int(duration[1]) * 60 - int(duration[2])
        else:
            self._begin_time = None

    def _write_sudoku(self, path):
        """Write the sudoku in a file."""
        csv_file = open(path, 'w', newline='')
        csv_writer = csv.writer(csv_file)
        for i in range(self._size_of_side):
            row = self._sudoku.GetSudokuGrid().GetRow(i)
            csv_writer.writerow(row)
        csv_file.close()

    def _write_type_of_sudoku(self, path):
        """Write the type of the sudoku in a file."""
        a_file = open(path, 'w')
        if self._x_sudoku:
            a_file.write('1')
        else:
            a_file.write('0')
        a_file.close()

    def _write_history_of_sudoku(self, path):
        """Write the history of the sudoku in a file."""
        a_file = open(path, 'w', newline='')
        csv_writer = csv.writer(a_file)
        csv_writer.writerow([self._sudoku.GetHistory().GetPosition()])
        for historyElem in self._sudoku.GetHistory().GetTrace():
            coord = historyElem.GetCoordinates()
            row = [coord[0], coord[1], historyElem.GetOldValue(),
                   historyElem.GetNewValue()]
            csv_writer.writerow(row)
        a_file.close()

    def _write_duration_of_sudoku(self, path):
        """Write the duration of the sudoku in a file."""
        a_file = open(path, 'w', newline='')
        csv_writer = csv.writer(a_file)
        csv_writer.writerow(self._duration)
        a_file.close()

    def _write_difficulty_of_sudoku(self, path):
        """Write the difficulty level of the sudoku in a file."""
        a_file = open(path, 'w')
        a_file.write(str(self._difficulty))
        a_file.close()

    def _unzip_game(self, path):
        """Unzip a saved game."""
        unzip_path = tempfile.mkdtemp()
        unzip_file, unzip_x, unzip_hist = None, None, None
        unzip_time, unzip_diff = None, None
        zfile = zipfile.ZipFile(path, 'r')
        for filename in zfile.namelist():
            unzip_filename = os.path.join(unzip_path, filename)
            if filename.lower().startswith(self._prefix['x']):
                unzip_x = unzip_filename
            elif filename.lower().startswith(self._prefix['h']):
                unzip_hist = unzip_filename
            elif filename.lower().startswith(self._prefix['t']):
                unzip_time = unzip_filename
            elif filename.lower().startswith(self._prefix['d']):
                unzip_diff = unzip_filename
            else:
                unzip_file = unzip_filename
            data = zfile.read(filename)
            a_file = open(unzip_filename, 'wb')
            a_file.write(data)
            a_file.close()
        return [unzip_path, unzip_file, unzip_x,
                unzip_hist, unzip_time, unzip_diff]

    def _zip_game(self, path, zip_path):
        """Save a game in a ZIP archive."""
        zfile = zipfile.ZipFile(path, 'w')
        os.chdir(zip_path)
        for a_file in os.listdir(zip_path):
            zfile.write(a_file)
        zfile.close()
        os.chdir(self._user_path)
        self._remove_temp_files(zip_path)

    def _remove_temp_files(self, temp_path):
        """Remove a temporary directory and the files it contains."""
        for a_file in os.listdir(temp_path):
            os.remove(os.path.join(temp_path, a_file))
        os.rmdir(temp_path)

    def _assign_events(self):
        """Assign the events to the buttons."""
        self.Bind(wx.EVT_MENU, self._on_new, id=wx.ID_NEW)
        self.Bind(wx.EVT_MENU, self._on_open, id=wx.ID_OPEN)
        self.Bind(wx.EVT_MENU, self._on_saveas, id=wx.ID_SAVEAS)
        self.Bind(wx.EVT_MENU, self._on_start_server, id=ID_SERVER)
        self.Bind(wx.EVT_MENU, self._on_join_server, id=ID_JOIN)
        self.Bind(wx.EVT_MENU, self._on_best, id=wx.ID_HIGHEST)
        self.Bind(wx.EVT_MENU, self._on_pause, id=ID_PAUSE)
        self.Bind(wx.EVT_MENU, self._on_solve, id=ID_STARTUP)
        self.Bind(wx.EVT_MENU, self._on_stop, id=wx.ID_STOP)
        self.Bind(wx.EVT_MENU, self._on_retry, id=wx.ID_RETRY)
        self.Bind(wx.EVT_MENU, self._on_undo, id=wx.ID_UNDO)
        self.Bind(wx.EVT_MENU, self._on_redo, id=wx.ID_REDO)
        self.Bind(wx.EVT_MENU, self._on_about, id=wx.ID_ABOUT)
        self.Bind(wx.EVT_MENU, self._on_exit, id=wx.ID_EXIT)
        self.Bind(wx.EVT_MENU, self._on_sounds, id=wx.ID_PREFERENCES)
        self.Bind(wx.EVT_MENU, self._on_change_port, id=ID_PORT)
        self.Bind(wx.EVT_MENU, self._on_change_init_colour, id=ID_INIT_COLOUR)
        self.Bind(wx.EVT_MENU, self._on_change_x_colour, id=ID_X_COLOUR)
        self.Bind(wx.EVT_MENU, self._on_change_bg_colour, id=ID_BG_COLOUR)
        self.Bind(wx.EVT_MENU, self._on_change_one_player_colour,
                  id=ID_ONE_P_COLOUR)
        self.Bind(wx.EVT_MENU, self._on_change_server_colour,
                  id=ID_SERV_P_COLOUR)
        self.Bind(wx.EVT_MENU, self._on_change_client_colour,
                  id=ID_CLI_P_COLOUR)
        self.Bind(wx.EVT_MENU, self._on_change_msg_colour, id=ID_MSG_COLOUR)
        self.Bind(wx.EVT_MENU, self._on_reset_colours, id=ID_RST_COLOURS)
        self.Bind(wx.EVT_MENU, self._on_forced_undo, id=ID_FORCED_UNDO)
        if wx.Platform != '__WXGTK__':
            self.Bind(wx.EVT_MENU, self._on_license, id=ID_LICENSE)
        self.Bind(wx.lib.intctrl.EVT_INT, self._on_int)
        self.Bind(wx.EVT_IDLE, self._on_idle)
        self.Bind(wx.EVT_BUTTON, self.OnSendMessage, id=ID_SEND)
        self.Bind(wx.EVT_TEXT_ENTER, self.OnSendMessage)
        self.Bind(wx.EVT_BUTTON, self._on_pause, id=ID_PAUSE)
        self.Bind(wx.EVT_CLOSE, self._on_exit)

    def _set_accelerators(self):
        """Set the accelerator table."""
        accel_list = [(wx.ACCEL_CTRL, ord('B'), wx.ID_HIGHEST),
                      (wx.ACCEL_CTRL, ord('Q'), wx.ID_EXIT),
                      (wx.ACCEL_CTRL, ord('N'), wx.ID_NEW),
                      (wx.ACCEL_CTRL, ord('E'), ID_SERVER),
                      (wx.ACCEL_CTRL, ord('J'), ID_JOIN),
                      (wx.ACCEL_CTRL, ord('O'), wx.ID_OPEN),
                      (wx.ACCEL_CTRL, ord('S'), wx.ID_SAVEAS),
                      (wx.ACCEL_CTRL, ord('R'), ID_STARTUP),
                      (wx.ACCEL_CTRL, ord('T'), wx.ID_STOP),
                      (wx.ACCEL_CTRL, ord('P'), ID_PAUSE),
                      (wx.ACCEL_CTRL, ord('G'), wx.ID_RETRY),
                      (wx.ACCEL_CTRL, ord('Z'), wx.ID_UNDO),
                      (wx.ACCEL_CTRL, ord('Y'), wx.ID_REDO),
                      (wx.ACCEL_CTRL, ord('M'), wx.ID_PREFERENCES),
                      (wx.ACCEL_CTRL, ord('W'), ID_PORT),
                      (wx.ACCEL_CTRL, ord('C'), ID_FORCED_UNDO),
                      (wx.ACCEL_CTRL, ord('F'), ID_INIT_COLOUR),
                      (wx.ACCEL_CTRL, ord('X'), ID_X_COLOUR),
                      (wx.ACCEL_CTRL, ord('D'), ID_BG_COLOUR),
                      (wx.ACCEL_CTRL, ord('0'), ID_ONE_P_COLOUR),
                      (wx.ACCEL_CTRL, ord('1'), ID_SERV_P_COLOUR),
                      (wx.ACCEL_CTRL, ord('2'), ID_CLI_P_COLOUR),
                      (wx.ACCEL_CTRL, ord('I'), ID_MSG_COLOUR),
                      (wx.ACCEL_CTRL, ord('V'), ID_RST_COLOURS),
                      (wx.ACCEL_CTRL, ord('A'), wx.ID_ABOUT),
                      (wx.ACCEL_ALT, ord(_('m')), ID_SEND)]
        if wx.Platform != '__WXGTK__':
            accel_list.append((wx.ACCEL_CTRL, ord('L'), ID_LICENSE))
        accel_table = wx.AcceleratorTable(accel_list)
        self.SetAcceleratorTable(accel_table)

    def _on_new(self, event):
        """Clear the Sudoku grid to start a new game."""
        if self._2_players and \
           (not self._is_server or self._sudoku_peer.IsThereAnyClient()):
            self._lock.acquire()
            self.SetClock()
            result = self._sudoku_peer.WriteIntoSocket('NEW', self._clock)
            self.DisplayMessage(
                         _("New game: waiting for the other player's answer."),
                         msg_colour)
            wx.YieldIfNeeded()
            self._lock.release()
            if result is None or self._sudoku_peer.IsPeerOk():
                self._start_new_game()
            self._lock.acquire()
            self.ResetClock()
            self._lock.release()
        else:
            self._start_new_game()

    def _on_open(self, event):
        """Open a ZIP file that contains a Sudoku puzzle."""
        file_selector = wx.FileDialog(self, _('Open the file'), self._user_path,
                                      wildcard=self._wildcard_label +
                                               self._wildcard_ext,
                                      style=wx.FD_OPEN | wx.FD_CHANGE_DIR)
        result = file_selector.ShowModal()
        path = file_selector.GetPath()
        a_file = file_selector.GetFilename()
        file_selector.Destroy()
        if result == wx.ID_OK and a_file != '':
            self.DestroyGrid()
            unzip_path, unzip_file, unzip_x, unzip_hist, \
                        unzip_time, unzip_diff = self._unzip_game(path)
            self._set_size_of_sudoku(unzip_file)
            self._set_type_of_sudoku(unzip_x)
            position, trace = self._set_history_of_sudoku(unzip_hist)
            self._set_duration_of_sudoku(unzip_time)
            self._set_difficulty_of_sudoku(unzip_diff)
            self.CreateGrid()
            self._display_grid(False, False)
            csv_file = open(unzip_file, 'r', newline='')
            csv_reader = csv.reader(csv_file)
            self._sudoku = SudokuPuzzle(self, self._size_of_side,
                                        self._x_sudoku, csv_reader,
                                        position, trace)
            csv_file.close()
            os.chdir(self._user_path)
            self._remove_temp_files(unzip_path)
            self._recolour_initial_grid()

    def _on_saveas(self, event):
        """Save a Sudoku puzzle in a ZIP file."""
        file_selector = wx.FileDialog(self, _('Save into the file'),
                                      self._user_path,
                                      wildcard=self._wildcard_label +
                                               self._wildcard_ext,
                                      style=wx.FD_SAVE |
                                            wx.FD_OVERWRITE_PROMPT |
                                            wx.FD_CHANGE_DIR)
        result = file_selector.ShowModal()
        path = file_selector.GetPath()
        a_file = file_selector.GetFilename()
        file_selector.Destroy()
        if result == wx.ID_OK and a_file != '':
            if not a_file.lower().endswith(self._file_ext):
                a_file += self._file_ext
                path += self._file_ext
            zip_path = tempfile.mkdtemp()
            filename = os.path.splitext(a_file)[0]
            self._write_sudoku(os.path.join(zip_path, filename))
            self._write_type_of_sudoku(os.path.join(zip_path,
                                                    self._prefix['x'] +
                                                    filename))
            self._write_history_of_sudoku(os.path.join(zip_path,
                                                       self._prefix['h'] +
                                                       filename))
            self._write_duration_of_sudoku(os.path.join(zip_path,
                                                        self._prefix['t'] +
                                                        filename))
            self._write_difficulty_of_sudoku(os.path.join(zip_path,
                                                          self._prefix['d'] +
                                                          filename))
            self._zip_game(path, zip_path)

    def _on_best(self, event):
        """Show the best scores."""
        self._show_best_scores(add_a_new_score=False)

    def _on_pause(self, event):
        """Pause the game."""
        if self._2_players:
            self._lock.acquire()
            self.SetClock()
            result = self._sudoku_peer.WriteIntoSocket('PSE', self._clock)
            self._lock.release()
            if result is None or self._sudoku_peer.IsPeerOk():
                self.PauseGame()
            self._lock.acquire()
            self.ResetClock()
            self._lock.release()
        else:
            self.PauseGame()

    def _on_solve(self, event):
        """Launch the resolution of the Sudoku puzzle."""
        if self._boxes[0].IsShown():
            try:
                self._sudoku.Solve()
            except StopError:
                pass
            except NoSolutionError:
                self._display_error_message(
                                    _('This Sudoku puzzle have no solution!'))

    def _on_stop(self, event):
        """Stop the resolution of the Sudoku puzzle."""
        if self._boxes[0].IsShown():
            self._sudoku.Stop()

    def _on_retry(self, event):
        """Retry the resolution from the beginning."""
        if self._boxes[0].IsShown() or self._button_pause is not None:
            if self._2_players and \
               (not self._is_server or self._sudoku_peer.IsThereAnyClient()):
                self._lock.acquire()
                self.SetClock()
                result = self._sudoku_peer.WriteIntoSocket('RET', self._clock)
                self.DisplayMessage(
                   _("Retry the game: waiting for the other player's answer."),
                   msg_colour)
                wx.YieldIfNeeded()
                self._lock.release()
                if result is None or self._sudoku_peer.IsPeerOk():
                    self._lock.acquire()
                    self._sudoku.GetHistory().Clear(True)
                    self.ResetInternalValues(False)
                    self.ShowGridPanel(True)
                    self._lock.release()
                self._lock.acquire()
                self.ResetClock()
                self._lock.release()
            else:
                self._sudoku.GetHistory().Clear(True)

    def _on_undo(self, event):
        """Undo the last operation."""
        if self._boxes[0].IsShown():
            if self._2_players and \
               (not self._is_server or self._sudoku_peer.IsThereAnyClient()):
                self._lock.acquire()
                elem = self._sudoku.GetHistory().GetCurrentHistoryElement()
                if elem is not None:
                    forcing_undo_allowed = False
                    if (elem.GetOldColour() != self._player_colour or \
                       elem.GetNewColour() != self._player_colour) and \
                       self._forced_undo:
                        self.SetClock()
                        result = self._sudoku_peer.WriteIntoSocket('AFU',
                                                                   self._clock)
                        self.DisplayMessage(
                        _("Force undo: waiting for the other player's answer."),
                        msg_colour)
                        wx.YieldIfNeeded()
                        self._lock.release()
                        if result is None or self._sudoku_peer.IsPeerOk():
                            forcing_undo_allowed = True
                        self._lock.acquire()
                        self.ResetClock()
                    if (elem.GetOldColour() == self._player_colour and \
                        elem.GetNewColour() == self._player_colour) or \
                       forcing_undo_allowed:
                        self._sudoku.GetHistory().MoveBack()
                        self._undo_number += 1
                        self._sudoku_peer.WriteIntoSocket('UND', 'UNDO')
                self._lock.release()
            else:
                self._sudoku.GetHistory().MoveBack()

    def _on_redo(self, event):
        """Redo the last undone operation."""
        if self._boxes[0].IsShown():
            if self._2_players:
                self._lock.acquire()
                if self._undo_number > 0:
                    self._sudoku.GetHistory().MoveOn()
                    self._undo_number -= 1
                    self._sudoku_peer.WriteIntoSocket('RED', 'REDO')
                self._lock.release()
            else:
                self._sudoku.GetHistory().MoveOn()

    def _on_about(self, event):
        """Display a message about the copyright and the license."""
        about_info = wx.adv.AboutDialogInfo()
        about_info.AddDeveloper(__author__)
        about_info.AddTranslator(__author__)
        about_info.SetCopyright(__copyright__)
        info = _('This software can generate and solve') + os.linesep
        info += _('different types of Sudoku puzzles like') + os.linesep
        info += _('those ones that can usually be found') + os.linesep
        info += _('in all kind of magazines these days.') + os.linesep
        info += os.linesep
        info += _('It is also possible to play this game') + os.linesep
        info += _('with someone else with the network mode.') + os.linesep
        about_info.SetDescription(info)
        if wx.Platform == '__WXGTK__':
            about_info.SetLicense(self._get_license())
        else:
            about_info.SetLicense(__license__)
        about_info.SetVersion('4.0.5')
        about_info.SetWebSite('http://yapysudoku.sourceforge.net')
        about_info.SetIcon(wx.Icon(os.path.join(self._path, 'icons',
                                   'yapysudoku.png'), wx.BITMAP_TYPE_PNG))
        wx.adv.AboutBox(about_info)

    def _on_license(self, event):
        """Display a message about the license."""
        my_license = self._get_license()
        about = wx.lib.dialogs.ScrolledMessageDialog(self, my_license,
                                                     self._license_menu_label,
                                                     size=(510, 300))
        about.ShowModal()
        about.Destroy()

    def _on_exit(self, event):
        """Quit the software."""
        if self._sudoku_peer is not None:
            self._sudoku_peer.stop()
        user_cfg_file = open(self._user_cfg_file, 'w')
        self._config_parser.write(user_cfg_file)
        user_cfg_file.close()
        self.Destroy()

    def _on_sounds(self, event):
        """Allow/disallow the sounds."""
        try:
            sounds_switch = event.GetEventObject().FindItemById(event.GetId())
            self._allow_sounds = sounds_switch.IsChecked()
            self._config_parser.set('Sounds', 'enabled',
                                    str(self._allow_sounds))
        except AttributeError:
            self._allow_sounds = event.IsChecked()

    def _on_forced_undo(self, event):
        """Allow/disallow the 'forced undo' in network mode."""
        try:
            forced_undo_switch = event.GetEventObject().FindItemById(
                                                                 event.GetId())
            self._forced_undo = forced_undo_switch.IsChecked()
            self._config_parser.set('Undo', 'forced',
                                    str(self._forced_undo))
        except AttributeError:
            self._allow_sounds = event.IsChecked()

    def _on_int(self, event):
        """
        When a number is put on the graphical grid, put this number
        on the grid of the SudokuPuzzle object.
        """
        if self._2_players:
            self._lock.acquire()
        self._cell = event.GetEventObject()
        if self._cell.GetColors()[0] == self._player_colour:
            cell_id = self._cell.GetId()
            row_number = cell_id // self._size_of_side
            column_number = cell_id % self._size_of_side
            value = self._cell.GetValue()
            old_value = self._sudoku.GetSudokuGrid().Get(row_number,
                                                         column_number)
            if old_value != value:
                self._try_to_put_a_value(old_value, value, row_number,
                                         column_number)
            elif self._2_players:
                self._lock.release()
        elif self._2_players:
            self._lock.release()

    def _on_idle(self, event):
        """
        - Set the focus when the GUI is on IDLE.
        - Display time.
        - Manage synchronisation problems between the GUI and the internal
          grid.
        - Manage network problems.
        """
        if self._focused_object is not None:
            self._focused_object.SetFocus()
            self._focused_object = None
        self._display_duration()
        if self._2_players:
            self._lock.acquire()
        self._manage_synchro_problems()
        if self._2_players:
            self._lock.release()
        if self._2_players:
            self._manage_network_problems()
        if self._begin_time is not None and self._count_time:
            wx.MilliSleep(10)
            wx.WakeUpIdle()

    def _on_start_server(self, event):
        """Start/stop the server."""
        if not self._2_players:
            self._set_2_players_mode(True)
            self._sudoku_peer = SudokuServer(self)
            self._sudoku_peer.start()
        else:
            self._sudoku_peer.stop()
            self._sudoku_peer = None
            self._set_1_player_mode()

    def _on_join_server(self, event):
        """Join a server or disconnect from a server."""
        if not self._2_players:
            server_choice = wx.TextEntryDialog(self, _('Enter server name:'),
                                        _('Server name'))
            result = server_choice.ShowModal()
            server = server_choice.GetValue()
            server_choice.Destroy()
            if result == wx.ID_OK and server != '':
                self._set_2_players_mode(False)
                self._sudoku_peer = SudokuClient(self, server)
                self._sudoku_peer.start()
        else:
            self._sudoku_peer.stop()
            self._sudoku_peer = None
            self._set_1_player_mode()

    def _on_change_port(self, event):
        """Change the port for network game."""
        global port
        choose_port = wx.TextEntryDialog(self, _('Enter port number:'),
                                         _('Port number'), str(port))
        result = choose_port.ShowModal()
        if result == wx.ID_OK:
            chosen_port = choose_port.GetValue()
            try:
                chosen_port = int(chosen_port)
            except ValueError:
                chosen_port = 0
            if chosen_port > MIN_PORT and chosen_port < MAX_PORT:
                port = chosen_port
                self._set_cfg_network_port()
        choose_port.Destroy()

    def _on_change_init_colour(self, event):
        """Change the foreground colour for initially filled cells."""
        global initially_filled_cell_colour
        colour_data = wx.ColourData()
        colour_data.SetColour(initially_filled_cell_colour)
        choose_colour = wx.ColourDialog(self, colour_data)
        result = choose_colour.ShowModal()
        if result == wx.ID_OK:
            chosen_colour = choose_colour.GetColourData().GetColour()
            if chosen_colour.IsOk() and chosen_colour != x_colour and \
               chosen_colour != bg_colour and \
               chosen_colour != one_player_colour and \
               chosen_colour != server_colour and \
               chosen_colour != client_colour:
                initially_filled_cell_colour = chosen_colour
                self._set_cfg_initially_filled_cell_colour()
        choose_colour.Destroy()

    def _on_change_x_colour(self, event):
        """Change the background colour for cells on diagonals in X game."""
        global x_colour
        colour_data = wx.ColourData()
        colour_data.SetColour(x_colour)
        choose_colour = wx.ColourDialog(self, colour_data)
        result = choose_colour.ShowModal()
        if result == wx.ID_OK:
            chosen_colour = choose_colour.GetColourData().GetColour()
            if chosen_colour.IsOk() and chosen_colour != bg_colour and \
               chosen_colour != initially_filled_cell_colour and \
               chosen_colour != one_player_colour and \
               chosen_colour != server_colour and \
               chosen_colour != client_colour:
                x_colour = chosen_colour
                self._set_cfg_diagonals_colour()
        choose_colour.Destroy()

    def _on_change_bg_colour(self, event):
        """Change the default background colour for cells."""
        global bg_colour
        colour_data = wx.ColourData()
        colour_data.SetColour(bg_colour)
        choose_colour = wx.ColourDialog(self, colour_data)
        result = choose_colour.ShowModal()
        if result == wx.ID_OK:
            chosen_colour = choose_colour.GetColourData().GetColour()
            if chosen_colour.IsOk() and chosen_colour != x_colour and \
               chosen_colour != initially_filled_cell_colour and \
               chosen_colour != one_player_colour and \
               chosen_colour != server_colour and \
               chosen_colour != client_colour:
                bg_colour = chosen_colour
                self._set_cfg_bg_colour()
        choose_colour.Destroy()

    def _on_change_one_player_colour(self, event):
        """Change player colour in one-player game."""
        global one_player_colour
        colour_data = wx.ColourData()
        colour_data.SetColour(one_player_colour)
        choose_colour = wx.ColourDialog(self, colour_data)
        result = choose_colour.ShowModal()
        if result == wx.ID_OK:
            chosen_colour = choose_colour.GetColourData().GetColour()
            if chosen_colour.IsOk() and chosen_colour != x_colour and \
               chosen_colour != bg_colour and \
               chosen_colour != initially_filled_cell_colour:
                one_player_colour = chosen_colour
                self._set_cfg_one_player_colour()
        choose_colour.Destroy()

    def _on_change_server_colour(self, event):
        """Change server colour in network game."""
        global server_colour
        colour_data = wx.ColourData()
        colour_data.SetColour(server_colour)
        choose_colour = wx.ColourDialog(self, colour_data)
        result = choose_colour.ShowModal()
        if result == wx.ID_OK:
            chosen_colour = choose_colour.GetColourData().GetColour()
            if chosen_colour.IsOk() and chosen_colour != x_colour and \
               chosen_colour != bg_colour and \
               chosen_colour != initially_filled_cell_colour and \
               chosen_colour != client_colour:
                server_colour = chosen_colour
                self._set_cfg_server_colour()
        choose_colour.Destroy()

    def _on_change_client_colour(self, event):
        """Change client colour in network game."""
        global client_colour
        colour_data = wx.ColourData()
        colour_data.SetColour(client_colour)
        choose_colour = wx.ColourDialog(self, colour_data)
        result = choose_colour.ShowModal()
        if result == wx.ID_OK:
            chosen_colour = choose_colour.GetColourData().GetColour()
            if chosen_colour.IsOk() and chosen_colour != x_colour and \
               chosen_colour != bg_colour and \
               chosen_colour != initially_filled_cell_colour and \
               chosen_colour != server_colour:
                client_colour = chosen_colour
                self._set_cfg_client_colour()
        choose_colour.Destroy()

    def _on_change_msg_colour(self, event):
        """Change text colour in network game."""
        global msg_colour
        colour_data = wx.ColourData()
        colour_data.SetColour(msg_colour)
        choose_colour = wx.ColourDialog(self, colour_data)
        result = choose_colour.ShowModal()
        if result == wx.ID_OK:
            chosen_colour = choose_colour.GetColourData().GetColour()
            if chosen_colour.IsOk():
                msg_colour = chosen_colour
                self._set_cfg_information_colour()
        choose_colour.Destroy()

    def _on_reset_colours(self, event):
        """Reset colors to their default values."""
        msg_dialog = wx.MessageDialog(self, _('Do you want to reset colours?'),
                                      _('Question'),
                                      style=wx.YES_NO | wx.ICON_QUESTION)
        result = msg_dialog.ShowModal()
        if result == wx.ID_YES:
            init_colours_with_default_values()
            self._create_cfg_colours_options()
        msg_dialog.Destroy()

    def _try_to_put_a_value(self, old_value, value, row_number, column_number):
        """Try to put a value into the internal sudoku grid."""
        if not self._sudoku.IsAPossibleValue(value, row_number, column_number):
            if self._2_players:
                self._lock.release()
            self.SetValue(old_value, row_number, column_number, to_peer=False)
            self._launch_error_wave_file()
            self._display_not_a_possible_value_error(value)
        else:
            if self._2_players:
                self._lock.release()
            self._sudoku.GetHistory().Push(row_number, column_number, old_value,
                                           value, True, self._player_colour,
                                           self._player_colour)
            if self._sudoku.GetSudokuGrid().GetBlanksNumber() == 0:
                self._display_end_game()

    def _get_license(self):
        """
        Read COPYING file to get the license.

        Return the license as a string.
        """
        result = ''
        copying = open(os.path.join(self._path, 'COPYING'), 'r')
        for line in copying:
            result += line
        copying.close()
        return result

    def _start_new_game(self):
        """Start a new game."""
        self.ResetInternalValues()
        self.DestroyGrid()
        self._choice_for_size_of_sudoku()
        self._choice_for_type_of_sudoku()
        self.CreateGrid()
        self._reset_sudoku()
        self.CreateNotEmptyGrid()
        self._config_parser.set('Game', 'size', str(self._size_of_side))
        self._config_parser.set('Game', 'normal', str(not self._x_sudoku))
        self._config_parser.set('Game', 'difficulty', str(self._difficulty))

    def _manage_synchro_problems(self):
        """
        Manage synchronisation problems between the GUI and the internal
        grid.
        """
        if self._begin_time is not None and self._count_time:
            if self._timer is None:
                self._timer = time.time()
            elif (self._size_of_box < 4 and time.time() - self._timer > 0.1) \
                 or time.time() - self._timer > 1:
                self._timer = None
                for i in range(self._size_of_side**2):
                    cell = self.FindWindowById(i)
                    value, gui_value, row_number, column_number = \
                                              self._get_values_and_coords(cell)
                    if gui_value != value:
                        self.SetValue(value, row_number, column_number,
                                      to_peer=False)

    def _manage_network_problems(self):
        """Manage network problems."""
        if self._network_problem.isSet():
            self._set_1_player_mode()
            self._network_problem.clear()
            self._sudoku_peer.stop()
            self._display_error_message(self._network_problem_msg)
            self._network_problem_msg = None

    def _get_values_and_coords(self, cell):
        """
        Get the values (internal and displayed)
        and the coordinates of a cell.
        """
        cell_id = cell.GetId()
        row_number = cell_id // self._size_of_side
        column_number = cell_id % self._size_of_side
        value = self._sudoku.GetSudokuGrid().Get(row_number,
                                                 column_number)
        gui_value = cell.GetValue()
        return [value, gui_value, row_number, column_number]

    def _set_2_players_mode(self, is_server):
        """Put the game in 2 players mode."""
        self._2_players = True
        self._is_server = is_server
        if is_server:
            self._player_colour = server_colour
        else:
            self._player_colour = client_colour
        self._display_msg.Clear()
        self._send_panel.Show(True)
        self._resize()
        self._recolour_grid()
        game_menu = self._remove_unneeded_buttons(is_server)
        if is_server:
            server_menu = game_menu.FindItemById(ID_SERVER)
            server_menu.SetItemLabel(self._stop_serv_menu_label +
                                     self._serv_ctrl_label)
            server_menu.SetHelp(self._stop_serv_long_label)
            game_menu.DestroyItem(ID_JOIN)
        else:
            client_menu = game_menu.FindItemById(ID_JOIN)
            client_menu.SetItemLabel(self._discon_menu_label +
                                     self._join_ctrl_label)
            client_menu.SetHelp(self._discon_long_label)
            game_menu.DestroyItem(ID_SERVER)

    def _set_1_player_mode(self):
        """Put the game in 1 player mode."""
        self._2_players = False
        self._player_colour = one_player_colour
        self._send_panel.Show(False)
        self._resize()
        self.ShowGridPanel(True)
        self._recolour_grid()
        self._sudoku.GetHistory().ResetColourFor1PlayerMode(self._player_colour)
        game_menu = self._readd_needed_buttons()
        if game_menu.FindItemById(ID_SERVER) is not None:
            server_menu = game_menu.FindItemById(ID_SERVER)
            server_menu.SetItemLabel(self._serv_menu_label +
                                     self._serv_ctrl_label)
            server_menu.SetHelp(self._serv_long_label)
            game_menu.Insert(2, ID_JOIN,
                             self._join_menu_label + self._join_ctrl_label,
                             self._join_long_label)
        else:
            client_menu = game_menu.FindItemById(ID_JOIN)
            client_menu.SetItemLabel(self._join_menu_label +
                                     self._join_ctrl_label)
            client_menu.SetHelp(self._join_long_label)
            game_menu.Insert(1, ID_SERVER,
                             self._serv_menu_label + self._serv_ctrl_label,
                             self._serv_long_label)

    def _remove_unneeded_buttons(self, is_server):
        """
        Remove unneeded buttons.

        Return the game menu.
        """
        game_menu = self._remove_unneeded_menu_buttons()
        toolbar = self._remove_unneeded_tool_buttons()
        if is_server:
            toolbar.InsertTool(4, ID_SERVER, self._stop_serv_long_label,
                               wx.Bitmap(os.path.join(self._icon_path,
                                                      'network-offline.png')),
                               wx.NullBitmap,
                               shortHelp=self._stop_serv_long_label,
                               longHelp=self._stop_serv_long_label)
        else:
            toolbar.InsertTool(4, ID_JOIN, self._discon_short_label,
                               wx.Bitmap(os.path.join(self._icon_path,
                                                      'network-offline.png')),
                               wx.NullBitmap,
                               shortHelp=self._discon_short_label,
                               longHelp=self._discon_long_label)
        toolbar.Realize()
        return game_menu

    def _remove_unneeded_menu_buttons(self):
        """
        Remove unneeded buttons in the game menu.

        Return the game menu.
        """
        game_menu = self.GetMenuBar().GetMenu(0)
        game_menu.DestroyItem(wx.ID_OPEN)
        game_menu.DestroyItem(wx.ID_SAVEAS)
        game_menu.DestroyItem(wx.ID_HIGHEST)
        game_menu.DestroyItem(ID_STARTUP)
        game_menu.DestroyItem(wx.ID_STOP)
        while len(self._menu_separators) > 0:
            sep = self._menu_separators.pop(0)
            game_menu.DestroyItem(sep)
        return game_menu

    def _remove_unneeded_tool_buttons(self):
        """
        Remove unneeded buttons in the tool bar.

        Return the tool bar.
        """
        toolbar = self.GetToolBar()
        toolbar.RemoveTool(wx.ID_OPEN)
        toolbar.RemoveTool(wx.ID_SAVEAS)
        toolbar.RemoveTool(ID_STARTUP)
        toolbar.RemoveTool(wx.ID_STOP)
        return toolbar

    def _readd_needed_buttons(self):
        """
        Readd needed buttons.

        Return the game menu.
        """
        game_menu = self._readd_needed_menu_buttons()
        toolbar = self._readd_needed_tool_buttons()
        toolbar.DeleteToolByPos(8)
        toolbar.Realize()
        return game_menu

    def _readd_needed_menu_buttons(self):
        """
        Readd needed buttons in the game menu.

        Return the game menu.
        """
        game_menu = self.GetMenuBar().GetMenu(0)
        game_menu.Insert(3, wx.ID_OPEN,
                         self._open_menu_label + self._open_ctrl_label,
                         self._open_long_label)
        game_menu.Insert(4, wx.ID_SAVEAS,
                         self._save_menu_label + self._save_ctrl_label,
                         self._save_long_label)
        self._menu_separators.append(game_menu.InsertSeparator(5))
        game_menu.Insert(6, wx.ID_HIGHEST,
                         self._best_menu_label + self._best_ctrl_label,
                         self._best_long_label)
        self._menu_separators.append(game_menu.InsertSeparator(7))
        game_menu.Insert(10, ID_STARTUP,
                         self._solve_menu_label + self._solve_ctrl_label,
                         self._solve_long_label)
        game_menu.Insert(11, wx.ID_STOP,
                         self._stop_menu_label + self._stop_ctrl_label,
                         self._stop_long_label)
        self._menu_separators.append(game_menu.InsertSeparator(12))
        return game_menu

    def _readd_needed_tool_buttons(self):
        """
        Readd needed buttons in the tool bar.

        Return the tool bar.
        """
        toolbar = self.GetToolBar()
        toolbar.InsertTool(1, wx.ID_OPEN, self._open_short_label,
                           wx.Bitmap(os.path.join(self._icon_path, 'open.png')),
                           wx.NullBitmap,
                           shortHelp=self._open_short_label,
                           longHelp=self._open_long_label)
        toolbar.InsertTool(2, wx.ID_SAVEAS, self._save_short_label,
                           wx.Bitmap(os.path.join(self._icon_path,
                                                  'save_as.png')),
                           wx.NullBitmap,
                           shortHelp=self._save_short_label,
                           longHelp=self._save_long_label)
        toolbar.InsertTool(6, ID_STARTUP,
                           self._solve_short_label,
                           wx.Bitmap(os.path.join(self._icon_path,
                                                  'solve.png')),
                           wx.NullBitmap,
                           shortHelp=self._solve_short_label,
                           longHelp=self._solve_long_label)
        toolbar.InsertTool(7, wx.ID_STOP, self._stop_short_label,
                           wx.Bitmap(os.path.join(self._icon_path, 'stop.png')),
                           wx.NullBitmap,
                           shortHelp=self._stop_short_label,
                           longHelp=self._stop_long_label)
        return toolbar

    def _modify_bars(self, pause):
        """Modify tool bar and menu bar wether game is paused or not."""
        if pause:
            pause_menu_label = self._unpause_menu_label
            pause_long_label = self._unpause_long_label
            pause_short_label = self._unpause_short_label
            icon = 'unpause.png'
        else:
            pause_menu_label = self._pause_menu_label
            pause_long_label = self._pause_long_label
            pause_short_label = self._pause_short_label
            icon = 'pause.png'
        pause_menu = self.GetMenuBar().GetMenu(0).FindItemById(ID_PAUSE)
        pause_menu.SetItemLabel(pause_menu_label + self._pause_ctrl_label)
        pause_menu.SetHelp(pause_long_label)
        toolbar = self.GetToolBar()
        toolbar.RemoveTool(ID_PAUSE)
        if self._2_players:
            toolbar.InsertTool(2, ID_PAUSE, pause_short_label,
                               wx.Bitmap(os.path.join(self._icon_path, icon)),
                               wx.NullBitmap,
                               shortHelp=pause_short_label,
                               longHelp=pause_long_label)
        else:
            toolbar.InsertTool(4, ID_PAUSE, pause_short_label,
                               wx.Bitmap(os.path.join(self._icon_path, icon)),
                               wx.NullBitmap,
                               shortHelp=pause_short_label,
                               longHelp=pause_long_label)
        toolbar.Realize()

    def _recolour_grid(self):
        """Change colour for cells filled by the player."""
        if self._boxes[0].IsShown() or self._button_pause is not None:
            for i in range(self._size_of_side**2):
                cell = self.FindWindowById(i)
                if cell.GetColors()[0] != initially_filled_cell_colour:
                    cell.SetColors(self._player_colour, self._player_colour)

    def _recolour_initial_grid(self):
        """Change colour for cells initially filled."""
        position = self._sudoku.GetHistory().GetPosition()
        while self._sudoku.GetHistory().MoveBack():
            pass
        for i in range(self._size_of_side**2):
            cell = self.FindWindowById(i)
            if cell.GetValue() is not None:
                cell.SetColors(initially_filled_cell_colour,
                               initially_filled_cell_colour)
            else:
                cell.SetColors(self._player_colour, self._player_colour)
        while self._sudoku.GetHistory().GetPosition() < position:
            self._sudoku.GetHistory().MoveOn()
        self._display_grid(True)
        self._boxes[0].SetFocus()
        if self._begin_time is not None:
            self._count_time = True

    def _display_error_message(self, msg):
        """Display an error message."""
        error = wx.MessageDialog(self, msg, self._error_label,
                                 style=wx.OK | wx.ICON_ERROR)
        error.ShowModal()
        error.Destroy()

    def _display_info_message(self, msg):
        """Display an information."""
        info = wx.MessageDialog(self, msg, _('Result'),
                                style=wx.OK | wx.ICON_INFORMATION)
        info.ShowModal()
        info.Destroy()

    def _display_not_a_possible_value_error(self, value):
        """Display a message indicating that the value is not possible."""
        msg = _('The number ') + str(value) + \
              _(' cannot be put at this place!')
        self._display_error_message(msg)

    def _display_end_game(self):
        """Display a message indicating that the game is ended."""
        self._set_duration(False)
        self._launch_success_wave_file()
        if self._2_players:
            self._display_2_players_result()
        else:
            user_scores_file = self._create_score_filename()
            scores = self._read_existing_scores(user_scores_file)
            if len(scores) == SCORE_SIZE:
                worst_score = scores[len(scores) - 1][:3]
            if len(scores) < SCORE_SIZE or self._duration < worst_score:
                winner = wx.TextEntryDialog(self, _('Enter your name:'),
                                            self._you_won,
                                            os.environ.get('LOGNAME', ''))
                winner.ShowModal()
                player_name = winner.GetValue()
                winner.Destroy()
                self._show_best_scores(player_name)
            else:
                self._display_info_message(self._you_won)
                self._show_best_scores(add_a_new_score=False)

    def _display_2_players_result(self):
        """
        Display a message indicating that
        the game is ended in 2 players mode.
        """
        cpt_you, cpt_other = 0, 0
        for i in range(self._size_of_side**2):
            cell = self.FindWindowById(i)
            if cell.GetColors()[0] == self._player_colour:
                cpt_you += 1
            elif cell.GetColors()[0] == self.GetOtherPlayerColour():
                cpt_other += 1
        msg = self._you + str(cpt_you) + self._filled_cells + os.linesep
        msg += self._other_player + str(cpt_other) + self._filled_cells
        msg += os.linesep + os.linesep
        if cpt_you > cpt_other:
            msg += self._you_won
        elif cpt_you < cpt_other:
            msg += _('Sorry, the other player won!')
        else:
            msg += _('There is no winner!')
        self._display_info_message(msg)

    def _launch_success_wave_file(self):
        """Launch the success wave file."""
        if self._allow_sounds:
            sound = pygame.mixer.Sound(os.path.join(self._path, 'sounds',
                                                    'success.wav'))
            sound.play()

    def _launch_error_wave_file(self):
        """Launch the error wave file."""
        if self._allow_sounds:
            sound = pygame.mixer.Sound(os.path.join(self._path, 'sounds',
                                                    'error.wav'))
            sound.play()

    def _read_existing_scores(self, user_scores_file):
        """
        Read existing scores.

        Return the list of scores.
        """
        result = []
        if os.access(user_scores_file, os.R_OK):
            csv_file = open(user_scores_file, 'r', newline='')
            csv_reader = csv.reader(csv_file)
            for row in csv_reader:
                result.append(row)
            csv_file.close()
        return result

    def _write_scores(self, user_scores_file, scores):
        """Write scores."""
        csv_file = open(user_scores_file, 'w', newline='')
        csv_writer = csv.writer(csv_file)
        for row in scores:
            csv_writer.writerow(row)
        csv_file.close()

    def _construct_scores_list(self, scores, duration=None):
        """Return a string representation of the list of scores."""
        result = ''
        i = 0
        for row in scores:
            i += 1
            if row == duration:
                result += '*****'
            result += str(i).rjust(2) + '   ' + self._duration_to_string(row,
                                                                         ':')
            result += ' (' + row[4] + ', ' + row[3] + ')'
            if row == duration:
                result += '*****'
            result += os.linesep
        return result

    def _create_score_filename(self):
        """
        Create the name of the file containing the scores.

        Return the name of the file containing the scores.
        """
        user_cfg_dir = self._create_cfg_dir()
        if not os.access(user_cfg_dir, os.F_OK):
            os.mkdir(user_cfg_dir)
        score_filename = 'scores' + str(self._size_of_side)
        if self._x_sudoku:
            score_filename += 'x'
        score_filename += str(self._difficulty)
        return os.path.join(user_cfg_dir, score_filename)

    def _show_best_scores(self, player_name=None, add_a_new_score=True):
        """Show the best scores at the end of a game."""
        user_scores_file = self._create_score_filename()
        scores = self._read_existing_scores(user_scores_file)
        if add_a_new_score:
            duration = [self._duration[0], self._duration[1],
                        self._duration[2], time.strftime('%c'), player_name]
            scores.append(duration)
            scores.sort()
            scores = scores[:SCORE_SIZE]
            self._write_scores(user_scores_file, scores)
            str_scores = self._construct_scores_list(scores, duration)
        else:
            str_scores = self._construct_scores_list(scores)
        winner = SudokuScrolledMessageDialog(self, str_scores,
                                             _('Best Scores'), size=(380, 236))
        winner.ShowModal()
        winner.Destroy()


class SudokuApp(wx.App):
    def OnInit(self):
        self.SetAppName('YAPySudoku')
        path = os.path.abspath(os.path.dirname(sys.argv[0]))
        localedir = os.path.join(path, 'locale')
        domain = 'yapysudoku'
        mylocale = wx.Locale(wx.LANGUAGE_DEFAULT)
        mylocale.AddCatalogLookupPathPrefix(localedir)
        mylocale.AddCatalog(domain)
        translation = gettext.translation(domain, localedir,
                                          [mylocale.GetCanonicalName()],
                                          fallback = True)
        translation.install()
        frame = SudokuFrame(path, self.GetAppName())
        self.SetTopWindow(frame)
        return True


def main():
    app = SudokuApp()
    if os.name == 'posix':
        mylocale = wx.Locale(wx.LANGUAGE_DEFAULT)
        encoding = str.upper(str(mylocale.GetSystemEncodingName()))
        try:
            if encoding == 'UTF-8' or encoding == 'UTF8':
                locale.setlocale(locale.LC_TIME, str(mylocale.GetCanonicalName()
                                                 + '.' + encoding))
            else:
                locale.setlocale(locale.LC_TIME,
                                 str(mylocale.GetCanonicalName()))
        except locale.Error:
            pass
    app.MainLoop()


if __name__ == '__main__':
    sys.exit(main())


